from decimal import Decimal

from chellow.e.bsuos import hh, neso_import
from chellow.e.computer import SupplySource
from chellow.models import (
    Comm,
    Contract,
    Cop,
    DtcMeterType,
    EnergisationStatus,
    GspGroup,
    MarketRole,
    MeterPaymentType,
    MeterType,
    Mtc,
    MtcLlfc,
    MtcParticipant,
    Participant,
    Pc,
    Site,
    Source,
    VoltageLevel,
    insert_comms,
    insert_cops,
    insert_dtc_meter_types,
    insert_energisation_statuses,
    insert_sources,
    insert_voltage_levels,
)
from chellow.utils import ct_datetime, to_utc, utc_datetime


def test_neso_import(mocker, sess):
    vf = to_utc(ct_datetime(1996, 1, 1))
    participant = Participant.insert(sess, "CALB", "AK Industries")
    market_role_Z = MarketRole.insert(sess, "Z", "Non-core")
    participant.insert_party(sess, market_role_Z, "None core", vf, None, None)
    sess.commit()

    log = mocker.Mock()
    set_progress = mocker.Mock()

    csv_file = []
    mocker.patch("chellow.e.bsuos.csv_latest", return_value=csv_file)

    neso_import(sess, log, set_progress)


def test_hh(sess):
    vf = to_utc(ct_datetime(1996, 1, 1))
    site = Site.insert(sess, "CI017", "Water Works")
    participant = Participant.insert(sess, "CALB", "Calb")
    market_role = MarketRole.insert(sess, "Z", "Non-core")
    participant.insert_party(sess, market_role, "None core", vf, None, None)
    ss_start = to_utc(ct_datetime(2025, 1, 1))
    Contract.insert_non_core(
        sess,
        "bsuos",
        "",
        {},
        ss_start,
        None,
        {"record": {"Fixed Tariff £/MWh": Decimal("6.22")}},
    )
    bank_holiday_rate_script = {"bank_holidays": []}
    Contract.insert_non_core(
        sess,
        "bank_holidays",
        "",
        {},
        utc_datetime(2000, 1, 1),
        None,
        bank_holiday_rate_script,
    )
    market_role_X = MarketRole.insert(sess, "X", "Supplier")
    market_role_M = MarketRole.insert(sess, "M", "Mop")
    market_role_C = MarketRole.insert(sess, "C", "HH Dc")
    market_role_R = MarketRole.insert(sess, "R", "Distributor")
    participant.insert_party(sess, market_role_M, "Fusion Mop", vf, None, None)
    participant.insert_party(sess, market_role_X, "Fusion", vf, None, None)
    participant.insert_party(sess, market_role_C, "Fusion DC", vf, None, None)
    mop_contract = Contract.insert_mop(
        sess, "Fusion", participant, "", {}, vf, None, {}
    )
    dc_contract = Contract.insert_dc(
        sess, "Fusion DC 2000", participant, "", {}, vf, None, {}
    )
    pc = Pc.insert(sess, "00", "hh", vf, None)
    insert_cops(sess)
    cop = Cop.get_by_code(sess, "5")
    insert_comms(sess)
    comm = Comm.get_by_code(sess, "GSM")
    imp_supplier_contract = Contract.insert_supplier(
        sess,
        "Fusion Supplier 2000",
        participant,
        "",
        {},
        utc_datetime(2000, 1, 1),
        None,
        {},
    )
    dno = participant.insert_party(sess, market_role_R, "WPD", vf, None, "22")
    Contract.insert_dno(
        sess,
        dno.dno_code,
        participant,
        "",
        {},
        vf,
        None,
        {},
    )
    meter_type = MeterType.insert(sess, "C5", "COP 1-5", vf, None)
    meter_payment_type = MeterPaymentType.insert(sess, "CR", "Credit", vf, None)
    mtc = Mtc.insert(sess, "845", False, True, vf, None)
    mtc_participant = MtcParticipant.insert(
        sess,
        mtc,
        participant,
        "HH COP5 And Above With Comms",
        False,
        True,
        meter_type,
        meter_payment_type,
        0,
        vf,
        None,
    )
    insert_voltage_levels(sess)
    voltage_level = VoltageLevel.get_by_code(sess, "HV")
    llfc = dno.insert_llfc(
        sess,
        "510",
        "PC 5-8 & HH HV",
        voltage_level,
        False,
        True,
        vf,
        None,
    )
    MtcLlfc.insert(sess, mtc_participant, llfc, vf, None)
    insert_sources(sess)
    source = Source.get_by_code(sess, "grid")
    insert_energisation_statuses(sess)
    energisation_status = EnergisationStatus.get_by_code(sess, "E")
    gsp_group = GspGroup.insert(sess, "_L", "South Western")
    insert_dtc_meter_types(sess)
    dtc_meter_type = DtcMeterType.get_by_code(sess, "H")
    supply = site.insert_e_supply(
        sess,
        source,
        None,
        "Bob",
        utc_datetime(2000, 1, 1),
        None,
        gsp_group,
        mop_contract,
        dc_contract,
        "hgjeyhuw",
        dno,
        pc,
        "845",
        cop,
        comm,
        None,
        energisation_status,
        dtc_meter_type,
        "22 7867 6232 781",
        "510",
        imp_supplier_contract,
        "7748",
        361,
        None,
        None,
        None,
        None,
        None,
    )
    sess.commit()
    caches = {}
    ss = SupplySource(
        sess,
        ss_start,
        ss_start,
        to_utc(ct_datetime(2021, 1, 1)),
        supply.eras[0],
        True,
        caches,
    )
    hh_datum = ss.hh_data[0]
    hh_datum["nbp-kwh"] = 1
    hh(ss)
    assert hh_datum["bsuos-rate"] == 0.00622



from chellow.e.lccc import run_import
from chellow.models import MarketRole, Participant
from chellow.utils import ct_datetime, to_utc


def test_lccc_import(mocker, sess):
    vf = to_utc(ct_datetime(1996, 1, 1))
    participant = Participant.insert(sess, "CALB", "AK Industries")
    market_role_Z = MarketRole.insert(sess, "Z", "Non-core")
    participant.insert_party(sess, market_role_Z, "None core", vf, None, None)
    sess.commit()
    mocker.patch("chellow.e.cfd.lccc_import", autospec=True)
    mocker.patch("chellow.e.rab.lccc_import", autospec=True)

    log = mocker.Mock()
    set_progress = mocker.Mock()

    run_import(sess, log, set_progress)



import csv
import sys
import threading
import traceback

from dateutil.relativedelta import relativedelta

from flask import g, redirect, render_template, request

from sqlalchemy.orm import joinedload
from sqlalchemy.sql.expression import false, select, true

from werkzeug.exceptions import BadRequest

from chellow.dloads import open_file
from chellow.e.issues import make_issue_bundles
from chellow.models import (
    Channel,
    Contract,
    Era,
    Issue,
    Party,
    RSession,
    Site,
    SiteEra,
    Snag,
    Supply,
    User,
)
from chellow.utils import (
    csv_make_val,
    hh_before,
    req_checkbox,
    req_date,
    req_int,
    req_int_none,
    req_str,
    utc_datetime_now,
)


def _make_rows(
    sess,
    now,
    contract,
    days_hidden,
    is_ignored,
    show_settlement,
    only_ongoing,
    days_long_hidden,
    comm_ids,
    limit=None,
):
    cutoff_date = now - relativedelta(days=days_hidden)
    q = (
        select(Snag, Channel, Era, Supply, SiteEra, Site, Contract)
        .join(Channel, Snag.channel_id == Channel.id)
        .join(Era, Channel.era_id == Era.id)
        .join(Supply, Era.supply_id == Supply.id)
        .join(SiteEra, Era.site_eras)
        .join(Site, SiteEra.site_id == Site.id)
        .join(Contract, Era.dc_contract_id == Contract.id)
        .where(
            SiteEra.is_physical == true(),
            Snag.start_date < cutoff_date,
            Era.comm_id.in_(comm_ids),
        )
        .order_by(
            Site.code,
            Supply.id,
            Channel.imp_related,
            Channel.channel_type,
            Snag.description,
            Snag.start_date,
            Snag.id,
        )
        .options(joinedload(Era.dc_contract))
    )
    if contract is not None:
        q = q.where(Era.dc_contract == contract)

    if not is_ignored:
        q = q.where(Snag.is_ignored == false())

    if show_settlement == "yes":
        q = q.join(Party, Supply.dno_id == Party.id).where(Party.dno_code != "99")
    elif show_settlement == "no":
        q = q.join(Party, Supply.dno_id == Party.id).where(Party.dno_code == "99")
    elif show_settlement == "both":
        pass
    else:
        raise BadRequest("show_settlement must be 'yes', 'no' or 'both'.")

    snag_groups = []
    prev_snag = None
    for snag, channel, era, supply, site_era, site, contract in sess.execute(q):
        snag_start = snag.start_date
        snag_finish = snag.finish_date

        if snag_finish is None:
            duration = now - snag_start
            age_of_snag = None
        else:
            duration = snag_finish - snag_start
            if hh_before(cutoff_date, snag_finish):
                age_of_snag = None
            else:
                delta = now - snag_finish
                age_of_snag = delta.days

        if only_ongoing and age_of_snag is not None:
            continue

        if days_long_hidden is not None and duration.days < days_long_hidden:
            continue

        if (
            prev_snag is None
            or channel.era != prev_snag.channel.era
            or snag.start_date != prev_snag.start_date
            or snag.finish_date != prev_snag.finish_date
            or snag.description != prev_snag.description
        ):
            if limit is not None and len(snag_groups) > limit:
                break

            issues = sess.scalars(
                select(Issue)
                .where(Issue.properties["supply_ids"].contains([supply.id]))
                .order_by(Issue.date_created)
            )
            issue_bundles = make_issue_bundles(sess, issues)

            snag_group = {
                "snags": [],
                "site": site,
                "era": era,
                "supply": supply,
                "description": snag.description,
                "start_date": snag.start_date,
                "finish_date": snag.finish_date,
                "contract": contract,
                "contract_name": contract.name,
                "hidden_days": days_hidden,
                "chellow_id": snag.id,
                "imp_mpan_core": era.imp_mpan_core,
                "exp_mpan_core": era.exp_mpan_core,
                "site_code": site.code,
                "site_name": site.name,
                "snag_description": snag.description,
                "channel_type": channel.channel_type,
                "is_ignored": snag.is_ignored,
                "days_since_finished": age_of_snag,
                "duration": duration.days,
                "issue_bundles": issue_bundles,
            }
            snag_groups.append(snag_group)
        snag_group["snags"].append(snag)
        prev_snag = snag

    def make_key(item):
        return item["duration"]

    snag_groups.sort(key=make_key, reverse=True)
    return snag_groups


def content(
    contract_id,
    days_hidden,
    is_ignored,
    user_id,
    only_ongoing,
    show_settlement,
    days_long_hidden,
    now,
    comm_ids,
):
    f = writer = None
    try:
        with RSession() as sess:
            user = User.get_by_id(sess, user_id)
            if contract_id is None:
                contract = None
                namef = "all"
            else:
                contract = Contract.get_dc_by_id(sess, contract_id)
                namef = contract.name

            f = open_file(f"channel_snags_{namef}.csv", user, mode="w", newline="")
            writer = csv.writer(f, lineterminator="\n")
            titles = (
                "contract",
                "hidden_days",
                "chellow_ids",
                "imp_mpan_core",
                "exp_mpan_core",
                "site_code",
                "site_name",
                "snag_description",
                "channel_types",
                "comm",
                "start_date",
                "finish_date",
                "is_ignored",
                "days_since_finished",
                "duration",
            )
            writer.writerow(titles)

            for snag_group in _make_rows(
                sess,
                now,
                contract,
                days_hidden,
                is_ignored,
                show_settlement,
                only_ongoing,
                days_long_hidden,
                comm_ids,
            ):

                vals = {
                    "contract": snag_group["contract"].name,
                    "hidden_days": days_hidden,
                    "chellow_ids": [snag.id for snag in snag_group["snags"]],
                    "imp_mpan_core": snag_group["imp_mpan_core"],
                    "exp_mpan_core": snag_group["exp_mpan_core"],
                    "site": snag_group["site"],
                    "site_code": snag_group["site"].code,
                    "site_name": snag_group["site"].name,
                    "snag_description": snag_group["description"],
                    "channel_types": [
                        f"{s.channel.imp_related}_{s.channel.channel_type}"
                        for s in snag_group["snags"]
                    ],
                    "start_date": snag_group["start_date"],
                    "finish_date": snag_group["finish_date"],
                    "is_ignored": snag_group["is_ignored"],
                    "days_since_finished": snag_group["days_since_finished"],
                    "duration": snag_group["duration"],
                    "comm": snag_group["era"].comm.code,
                }

                writer.writerow(csv_make_val(vals[t]) for t in titles)
    except BaseException:
        msg = traceback.format_exc()
        sys.stderr.write(msg)
        writer.writerow([msg])
    finally:
        if f is not None:
            f.close()


LIMIT = 200


def do_get(sess):
    contract_id = req_int_none("dc_contract_id")
    days_hidden = req_int("days_hidden")
    is_ignored = req_checkbox("is_ignored")
    only_ongoing = req_checkbox("only_ongoing")
    show_settlement = req_str("show_settlement")
    as_csv = req_checkbox("as_csv")
    days_long_hidden = req_int_none("days_long_hidden")
    if "now_year" in request.values:
        now = req_date("now")
    else:
        now = utc_datetime_now()
    comm_ids = [int(x) for x in request.values.getlist("comm_id")]

    if as_csv:
        args = (
            contract_id,
            days_hidden,
            is_ignored,
            g.user.id,
            only_ongoing,
            show_settlement,
            days_long_hidden,
            now,
            comm_ids,
        )
        threading.Thread(target=content, args=args).start()
        return redirect("/downloads", 303)
    else:
        if contract_id is None:
            contract = None
        else:
            contract = Contract.get_dc_by_id(sess, contract_id)
        now = utc_datetime_now()
        snag_groups = _make_rows(
            sess,
            now,
            contract,
            days_hidden,
            is_ignored,
            show_settlement,
            only_ongoing,
            days_long_hidden,
            comm_ids,
            limit=LIMIT,
        )
        return render_template(
            "reports/channel_snags.html",
            contract=contract,
            limit=LIMIT,
            snag_groups=snag_groups,
            days_hidden=days_hidden,
            is_ignored=is_ignored,
            only_ongoing=only_ongoing,
            show_settlement=show_settlement,
            days_long_hidden="" if days_long_hidden is None else days_long_hidden,
            comm_ids=comm_ids,
        )



from datetime import datetime as Datetime

from decimal import Decimal, InvalidOperation
from openpyxl import load_workbook

from werkzeug.exceptions import BadRequest

from chellow.utils import HH, ct_datetime, hh_before, hh_range, to_utc


def create_parser(input_stream, mpan_map, messages):
    return HhParserVital(input_stream, mpan_map, messages)


def get_cell(sheet, col, row):
    try:
        coordinates = f"{col}{row}"
        return sheet[coordinates]
    except IndexError:
        raise BadRequest(f"Can't find the cell {coordinates} on sheet {sheet}.")


def get_date(sheet, col, row):
    cell = get_cell(sheet, col, row)
    val = cell.value
    if not isinstance(val, Datetime):
        raise BadRequest(
            f"Problem reading {val} (of type {type(val)}) as a timestamp at "
            f"{cell.coordinate}."
        )
    return val


def get_str(sheet, col, row):
    return get_cell(sheet, col, row).value.strip()


def get_dec(sheet, col, row):
    return get_dec_from_cell(get_cell(sheet, col, row))


def get_dec_from_cell(cell):
    try:
        return Decimal(str(cell.value))
    except InvalidOperation as e:
        raise BadRequest(f"Problem parsing the number at {cell.coordinate}. {e}")


def get_int(sheet, col, row):
    return int(get_cell(sheet, col, row).value)


def find_hhs(sheet, imp_mpan, exp_mpan):
    row = pres_readings = pres_ts = None
    caches = {}
    try:
        for row in range(2, len(sheet["A"]) + 1):
            imp_read_cell = get_cell(sheet, "H", row)
            imp_read_val = imp_read_cell.value
            if imp_read_val is None or imp_read_val == "":
                break

            exp_read_cell = get_cell(sheet, "I", row)
            exp_read_val = exp_read_cell.value
            if exp_read_val is None or exp_read_val == "":
                break

            # '16/09/2024 14:45:00
            ts_str = get_str(sheet, "A", row)
            ts_naive = Datetime.strptime(ts_str, "%d/%m/%Y %H:%M:%S")
            ts_ct = ct_datetime(
                ts_naive.year,
                ts_naive.month,
                ts_naive.day,
                ts_naive.hour,
                ts_naive.minute,
            )

            imp_read = get_dec_from_cell(imp_read_cell)
            exp_read = get_dec_from_cell(exp_read_cell)
            ts = to_utc(ts_ct)

            if ts_naive.minute in (0, 30) and hh_before(ts, pres_ts):
                if pres_readings is not None:
                    pres_imp_read, pres_exp_read = pres_readings
                    hhs = hh_range(caches, ts, pres_ts - HH)
                    num_hhs = len(hhs)
                    val_imp = (pres_imp_read - imp_read) / num_hhs
                    val_exp = (pres_exp_read - exp_read) / num_hhs
                    for hh in hhs:

                        yield {
                            "mpan_core": imp_mpan,
                            "channel_type": "ACTIVE",
                            "start_date": hh,
                            "value": val_imp,
                            "status": "A",
                        }
                        yield {
                            "mpan_core": exp_mpan,
                            "channel_type": "ACTIVE",
                            "start_date": hh,
                            "value": val_exp,
                            "status": "A",
                        }

                pres_readings = imp_read, exp_read
                pres_ts = ts

    except BadRequest as e:
        e.description = (f"Problem at row: {row}: {e.description}",)
        raise e


class HhParserVital:
    def __init__(self, input_stream, mpan_map, messages):
        book = load_workbook(input_stream, data_only=True)
        sheet = book.worksheets[0]
        imp_name = get_str(sheet, "H", 1).strip()
        imp_mpan = mpan_map[imp_name]
        exp_name = get_str(sheet, "I", 1).strip()
        exp_mpan = mpan_map[exp_name]
        self.data = iter(find_hhs(sheet, imp_mpan, exp_mpan))

    def __iter__(self):
        return self

    def __next__(self):
        return next(self.data)

    def close(self):
        pass



from pathlib import Path
from tempfile import TemporaryDirectory

from chellow.dloads import SERIAL_DIGITS, open_file, startup


def test_startup():
    with TemporaryDirectory() as tmpdir:
        startup(Path(tmpdir), run_deleter=False)


def test_startup_interrupted():
    with TemporaryDirectory() as tempdir:
        tmpdir = Path(tempdir)
        downloads = tmpdir / "downloads"
        downloads.mkdir()
        fname = "0" * SERIAL_DIGITS + "_RUNNING_inewton.ods"
        (downloads / fname).touch()
        startup(tmpdir, run_deleter=False)


def test_truncate():
    with TemporaryDirectory() as tmpdir:
        startup(Path(tmpdir), run_deleter=False)
        user = None
        with open_file("test.txt", user, mode="w") as fl:
            fl.truncate()



from sqlalchemy import select


from chellow.e.neso import csv_latest, parse_date
from chellow.models import Contract, RateScript
from chellow.utils import (
    ct_datetime,
    to_utc,
)

BANDED_START = to_utc(ct_datetime(2023, 4, 1))


def hh(ds):
    if "tnuos" not in ds.non_primary_elements:
        for hh in ds.hh_data:
            if hh["start-date"] >= BANDED_START:
                _process_banded_hh(ds, hh)


BAND_LOOKUP = {
    "Domestic Aggregated (Related MPAN)": "Domestic",
    "Domestic Aggregated with Residual": "Domestic",
    "Domestic Aggregated or CT with Residual": "Domestic",
    "HV Generation Site Specific": "HV1",
    "HV Generation Site Specific no RP charge": "HV1",
    "HV Site Specific Band 1": "HV1",
    "HV Site Specific Band 2": "HV2",
    "HV Site Specific Band 3": "HV3",
    "HV Site Specific Band 4": "HV4",
    "HV Site Specific No Residual": "HV1",
    "LV Generation Aggregated": "LV1",
    "LV Generation Site Specific": "LV1",
    "LV Generation Site Specific no RP charge": "LV1",
    "LV Site Specific Band 1": "LV1",
    "LV Site Specific Band 2": "LV2",
    "LV Site Specific Band 3": "LV3",
    "LV Site Specific Band 4": "LV4",
    "LV Site Specific No Residual": "LV1",
    "LV Sub Generation Aggregated": "LV1",
    "LV Sub Generation Site Specific": "LV1",
    "LV Sub Generation Site Specific no RP charge": "LV1",
    "LV Sub Site Specific Band 1": "LV1",
    "LV Sub Site Specific Band 2": "LV2",
    "LV Sub Site Specific Band 3": "LV3",
    "LV Sub Site Specific Band 4": "LV4",
    "LV Sub Site Specific No Residual": "LV1",
    "Non-Domestic Aggregated (related MPAN)": "LV_NoMIC_1",
    "Non-Domestic Aggregated Band 1": "LV_NoMIC_1",
    "Non-Domestic Aggregated or CT Band 1": "LV_NoMIC_1",
    "Non-Domestic Aggregated Band 2": "LV_NoMIC_2",
    "Non-Domestic Aggregated or CT Band 2": "LV_NoMIC_2",
    "Non-Domestic Aggregated Band 3": "LV_NoMIC_3",
    "Non-Domestic Aggregated or CT Band 3": "LV_NoMIC_3",
    "Non-Domestic Aggregated Band 4": "LV_NoMIC_4",
    "Non-Domestic Aggregated or CT Band 4": "LV_NoMIC_4",
    "Non-Domestic Aggregated No Residual": "LV_NoMIC_1",
    "Non-Domestic Aggregated or CT No Residual": "LV_NoMIC_1",
    "Unmetered Supplies": "Unmetered",
    "Designated EHV Site Specific No Residual": "EHV1",
    "Designated EHV Site Specific Band 1": "EHV1",
    "Designated EHV Site Specific Band 2": "EHV2",
    "Designated EHV Site Specific Band 3": "EHV3",
    "Designated EHV Site Specific Band 4": "EHV4",
    "Designated EHV0": "EHV1",
    "Designated EHV1": "EHV1",
    "Designated EHV2": "EHV2",
    "Designated EHV3": "EHV3",
    "Designated EHV4": "EHV4",
}


def _process_banded_hh(ds, hh):
    rates = ds.non_core_rate("tnuos", hh["start-date"])
    band_code = BAND_LOOKUP[hh["duos-description"]]
    hh["tnuos-band"] = band_code
    if hh["ct-decimal-hour"] == 12:
        rate = float(rates["bands"][band_code]["TDR Tariff"])
        hh["tnuos-rate"] = rate
        if band_code == "Unmetered":
            hh["tnuos-gbp"] = rate / 100 * ds.sc / 365
        else:
            hh["tnuos-gbp"] = rate
        hh["tnuos-days"] = 1


def neso_import(sess, log, set_progress):
    log("Starting to check for new TNUoS TDR Tariffs")

    contract_name = "tnuos"
    contract = Contract.find_non_core_by_name(sess, contract_name)
    if contract is None:
        contract = Contract.insert_non_core(
            sess,
            contract_name,
            "",
            {"enabled": True},
            to_utc(ct_datetime(1996, 4, 1)),
            None,
            {},
        )
        sess.commit()
    state = contract.make_state()
    last_import_date = state.get("last_import_date")

    for record in csv_latest(
        "transmission-network-use-of-system-tnuos-tariffs",
        last_import_date,
        name="transmission_demand_residual_(tdr)_tariffs",
    ):
        # {
        #   "_id": 1,
        #   "Publication": "Final",
        #   "Year_FY": 2024,
        #   "Published_Date": "2023-02-14",
        #   "TDR Band": "Domestic",
        #   "TDR Tariff": 0.119264,
        #   "Notes": "Domestic"
        # },
        fy_year = int(record["Year_FY"]) - 1
        fy_start = to_utc(ct_datetime(fy_year, 4, 1))
        if fy_start < contract.start_rate_script.start_date:
            continue

        rs = sess.execute(
            select(RateScript).where(
                RateScript.contract == contract,
                RateScript.start_date == fy_start,
            )
        ).scalar_one_or_none()
        if rs is None:
            rs = contract.insert_rate_script(sess, fy_start, {})

        rs_script = rs.make_script()
        try:
            bands = rs_script["bands"]
        except KeyError:
            bands = rs_script["bands"] = {}

        record_key = record["TDR Band"]
        record_published_date = parse_date(record["Published_Date"])

        band = bands.get(record_key)
        if band is None or parse_date(band["Published_Date"]) < record_published_date:
            if "TDR Tariff" not in record:
                tdr_tariff = [
                    v for k, v in record.items() if k.startswith("TDR Tariff")
                ][0]
                record["TDR Tariff"] = tdr_tariff
            bands[record_key] = record
            rs.update(rs_script)
            sess.commit()

    log("Finished TNUoS TDR Tariffs")
    sess.commit()



import csv
from collections import defaultdict
from datetime import datetime as Datetime
from decimal import Decimal
from importlib import import_module
from io import BytesIO, StringIO


from dateutil.relativedelta import relativedelta

from flask import (
    Blueprint,
    flash,
    g,
    make_response,
    redirect,
    render_template as rtemplate,
    request,
)

from sqlalchemy import Integer, cast, false, func, null, select, text, true
from sqlalchemy.orm import aliased, joinedload


from werkzeug.exceptions import BadRequest

from zish import dumps, loads

import chellow.gas.bill_import
import chellow.gas.dn_rate_parser
from chellow.gas.issues import make_issue_bundle, make_issue_bundles
from chellow.models import (
    BillType,
    Contract,
    GBatch,
    GBill,
    GContract,
    GDn,
    GEra,
    GExitZone,
    GIssue,
    GIssueEntry,
    GLdz,
    GRateScript,
    GReadType,
    GReadingFrequency,
    GRegisterRead,
    GSupply,
    GUnit,
    Report,
    Site,
    SiteGEra,
    User,
    UserRole,
)
from chellow.utils import (
    HH,
    csv_make_val,
    req_checkbox,
    req_date,
    req_decimal,
    req_file,
    req_int,
    req_markdown,
    req_str,
    req_zish,
    utc_datetime,
    utc_datetime_now,
)
from chellow.views import hx_redirect as chx_redirect


def chellow_redirect(path, code=None):
    return redirect(f"/g{path}", code)


def hx_redirect(path, code=None):
    return chx_redirect(f"/g{path}", code)


gas = Blueprint("g", __name__, template_folder="templates", url_prefix="/g")


def render_template(tname, **kwargs):
    return rtemplate(f"g/{tname}", **kwargs)


@gas.route("/dashboard", methods=["GET"])
def dashboard():
    return render_template("dashboard.html")


@gas.route("/issues")
def issues_get():
    supplier_contracts = g.sess.scalars(
        select(GContract)
        .join(GIssue)
        .where(GContract.is_industry == false())
        .distinct()
        .order_by(GContract.name)
    ).all()
    owners = g.sess.scalars(
        select(User)
        .join(GIssue, User.id == cast(GIssue.properties["owner_id"], Integer))
        .distinct()
        .order_by(User.email_address)
    ).all()
    return render_template(
        "issues.html", supplier_contracts=supplier_contracts, owners=owners
    )


@gas.route("/supplier_contracts/<int:contract_id>/issues")
def supplier_issues_get(contract_id):
    contract = GContract.get_supplier_by_id(g.sess, contract_id)
    issues = g.sess.scalars(
        select(GIssue)
        .where(GIssue.g_contract == contract)
        .order_by(GIssue.is_open.desc(), GIssue.date_created)
    )
    bundles = make_issue_bundles(g.sess, issues)
    return render_template(
        "supplier_issues.html", contract=contract, issue_bundles=bundles
    )


@gas.route("/supplier_contracts/<int:contract_id>/add_issue")
def issue_add_get(contract_id):
    contract = GContract.get_supplier_by_id(g.sess, contract_id)
    if "supply_id" in request.values:
        supply_id = req_int("supply_id")
        supply = GSupply.get_by_id(g.sess, supply_id)
    else:
        supply = None

    return render_template("issue_add.html", contract=contract, supply=supply)


@gas.route("/supplier_contracts/<int:contract_id>/add_issue", methods=["POST"])
def issue_add_post(contract_id):
    contract = GContract.get_supplier_by_id(g.sess, contract_id)
    try:
        subject = req_str("subject")

        now = utc_datetime_now()
        properties = {"subject": subject, "owner_id": g.user.id}
        if "supply_id" in request.values:
            properties["supply_ids"] = [req_int("supply_id")]
        issue = contract.insert_issue(g.sess, now, properties)
        g.sess.commit()
        return chellow_redirect(f"/issues/{issue.id}", 303)
    except BadRequest as e:
        flash(e.description)
        return make_response(render_template("issue_add.html", contract=contract), 400)


@gas.route("/issues/<int:issue_id>")
def issue_get(issue_id):
    issue = GIssue.get_by_id(g.sess, issue_id)
    issue_bundle = make_issue_bundle(g.sess, issue)
    return render_template("issue.html", issue_bundle=issue_bundle)


@gas.route("/issues/<int:issue_id>/edit")
def issue_edit_get(issue_id):
    issue = GIssue.get_by_id(g.sess, issue_id)
    users = g.sess.scalars(
        select(User)
        .join(UserRole)
        .where(UserRole.code == "editor")
        .order_by(User.email_address)
    )
    return render_template("issue_edit.html", issue=issue, users=users)


@gas.route("/issues/<int:issue_id>/edit", methods=["POST"])
def issue_edit_post(issue_id):
    try:
        issue = GIssue.get_by_id(g.sess, issue_id)
        date_created = req_date("date_created")
        is_open = req_checkbox("is_open")
        owner_id = req_int("owner_id")
        subject = req_str("subject")
        props = issue.properties
        props["owner_id"] = owner_id
        props["subject"] = subject
        issue.update(date_created, is_open, props)
        g.sess.commit()
        return chellow_redirect(f"/issues/{issue.id}", 303)
    except BadRequest as e:
        flash(e.description)
        return make_response(render_template("issue_edit.html", issue=issue), 400)


@gas.route("/issues/<int:issue_id>/edit", methods=["DELETE"])
def issue_edit_delete(issue_id):
    try:
        issue = GIssue.get_by_id(g.sess, issue_id)
        contract = issue.g_contract
        issue.delete(g.sess)
        g.sess.commit()
        return hx_redirect(f"/supplier_contracts/{contract.id}/issues", 303)
    except BadRequest as e:
        flash(e.description)
        return make_response(render_template("issue_edit.html", issue=issue), 400)


@gas.route("/issues/<int:issue_id>/attach_supply")
def issue_attach_supply_get(issue_id):
    issue = GIssue.get_by_id(g.sess, issue_id)
    return render_template("issue_attach_supply.html", issue=issue)


@gas.route("/issues/<int:issue_id>/attach_supply", methods=["POST"])
def issue_attach_supply_post(issue_id):
    issue = GIssue.get_by_id(g.sess, issue_id)
    try:
        mprn = req_str("mprn")
        supply = GSupply.get_by_mprn(g.sess, mprn)
        issue.attach_g_supply(supply)
        g.sess.commit()
        return chellow_redirect(f"/issues/{issue.id}", 303)
    except BadRequest as e:
        flash(e.description)
        return make_response(
            render_template("issue_attach_supply.html", issue=issue), 400
        )


@gas.route("/issues/<int:issue_id>/add_entry")
def entry_add_get(issue_id):
    issue = GIssue.get_by_id(g.sess, issue_id)
    return render_template("entry_add.html", issue=issue)


@gas.route("/issues/<int:issue_id>/add_entry", methods=["POST"])
def entry_add_post(issue_id):
    issue = GIssue.get_by_id(g.sess, issue_id)
    try:
        markdown = req_markdown("markdown")

        issue.add_g_entry(g.sess, markdown)
        g.sess.commit()
        return chellow_redirect(f"/issues/{issue.id}", 303)
    except BadRequest as e:
        flash(e.description)
        return make_response(render_template("entry_add.html", issue=issue), 400)


@gas.route("/entries/<int:entry_id>/edit")
def entry_edit_get(entry_id):
    entry = GIssueEntry.get_by_id(g.sess, entry_id)
    return render_template("entry_edit.html", entry=entry)


@gas.route("/entries/<int:entry_id>/edit", methods=["POST"])
def entry_edit_post(entry_id):
    entry = GIssueEntry.get_by_id(g.sess, entry_id)
    try:
        markdown = req_markdown("markdown")
        timestamp = req_date("timestamp")

        entry.update(timestamp, markdown)
        g.sess.commit()
        return chellow_redirect(f"/issues/{entry.g_issue.id}", 303)
    except BadRequest as e:
        flash(e.description)
        return make_response(render_template("entry_edit.html", entry=entry), 400)


@gas.route("/entries/<int:entry_id>/edit", methods=["DELETE"])
def entry_edit_delete(entry_id):
    try:
        entry = GIssueEntry.get_by_id(g.sess, entry_id)
        issue = entry.issue
        entry.delete(g.sess)
        g.sess.commit()
        return hx_redirect(f"/issues/{issue.id}", 303)
    except BadRequest as e:
        flash(e.description)
        return make_response(render_template("entry_edit.html", entry=entry), 400)


@gas.route("/supplies")
def supplies_get():
    if "search_pattern" in request.values:
        pattern = req_str("search_pattern")
        pattern = pattern.strip()
        reduced_pattern = pattern.replace(" ", "")
        if "max_results" in request.values:
            max_results = req_int("max_results")
        else:
            max_results = 50

        g_eras = (
            g.sess.query(GEra)
            .from_statement(
                text(
                    "select e1.* from g_era as e1 "
                    "inner join (select e2.g_supply_id, max(e2.start_date) "
                    "as max_start_date from g_era as e2 "
                    "join g_supply on (e2.g_supply_id = g_supply.id) "
                    "where replace(lower(g_supply.mprn), ' ', '') "
                    "like lower(:reducedPattern) "
                    "or lower(e2.account) like lower(:pattern) "
                    "or lower(e2.msn) like lower(:pattern) "
                    "group by e2.g_supply_id) as sq "
                    "on e1.g_supply_id = sq.g_supply_id "
                    "and e1.start_date = sq.max_start_date limit :max_results"
                )
            )
            .params(
                pattern="%" + pattern + "%",
                reducedPattern="%" + reduced_pattern + "%",
                max_results=max_results,
            )
            .all()
        )
        if len(g_eras) == 1:
            return chellow_redirect(f"/supplies/{g_eras[0].g_supply.id}")
        else:
            return render_template(
                "supplies.html", g_eras=g_eras, max_results=max_results
            )
    else:
        return render_template("supplies.html")


@gas.route("/supplies/<int:g_supply_id>")
def supply_get(g_supply_id):
    debug = ""
    try:
        g_era_bundles = []
        g_supply = GSupply.get_by_id(g.sess, g_supply_id)
        g_eras = g.sess.scalars(
            select(GEra)
            .where(GEra.g_supply == g_supply)
            .order_by(GEra.start_date.desc())
        ).all()
        for g_era in g_eras:
            physical_site = (
                g.sess.query(Site)
                .join(SiteGEra)
                .filter(SiteGEra.is_physical == true(), SiteGEra.g_era == g_era)
                .one()
            )
            other_sites = (
                g.sess.query(Site)
                .join(SiteGEra)
                .filter(SiteGEra.is_physical != true(), SiteGEra.g_era == g_era)
                .all()
            )
            g_bill_dicts = []
            g_era_bundle = {
                "g_era": g_era,
                "physical_site": physical_site,
                "other_sites": other_sites,
                "g_bill_dicts": g_bill_dicts,
            }
            g_era_bundles.append(g_era_bundle)

            g_era_bundle["shared_accounts"] = (
                g.sess.query(GSupply)
                .distinct()
                .join(GEra)
                .filter(
                    GSupply.id != g_supply.id,
                    GEra.account == g_era.account,
                    GEra.g_contract == g_era.g_contract,
                )
                .all()
            )

            g_bills = (
                select(GBill)
                .where(GBill.g_supply == g_supply)
                .order_by(
                    GBill.start_date.desc(),
                    GBill.issue_date.desc(),
                    GBill.reference.desc(),
                )
            )
            if g_era != g_eras[0]:
                g_bills = g_bills.where(GBill.start_date < g_era.finish_date)
            if g_era != g_eras[-1]:
                g_bills = g_bills.where(GBill.start_date >= g_era.start_date)

            for g_bill in g.sess.scalars(g_bills):
                g_reads = (
                    g.sess.query(GRegisterRead)
                    .filter(GRegisterRead.g_bill == g_bill)
                    .order_by(GRegisterRead.pres_date.desc())
                    .options(
                        joinedload(GRegisterRead.prev_type),
                        joinedload(GRegisterRead.pres_type),
                    )
                    .all()
                )
                g_bill_dicts.append({"g_bill": g_bill, "g_reads": g_reads})

            b_dicts = list(reversed(g_bill_dicts))
            for i, b_dict in enumerate(b_dicts):
                if i < (len(b_dicts) - 1):
                    g_bill = b_dict["g_bill"]
                    next_b_dict = b_dicts[i + 1]
                    next_g_bill = next_b_dict["g_bill"]
                    if (
                        g_bill.start_date,
                        g_bill.finish_date,
                        g_bill.kwh,
                        g_bill.net,
                    ) == (
                        next_g_bill.start_date,
                        next_g_bill.finish_date,
                        -1 * next_g_bill.kwh,
                        -1 * next_g_bill.net,
                    ) and "collapsible" not in b_dict:
                        b_dict["collapsible"] = True
                        next_b_dict["first_collapsible"] = True
                        next_b_dict["collapsible"] = True
                        b_dict["collapse_id"] = next_b_dict["collapse_id"] = g_bill.id

        RELATIVE_YEAR = relativedelta(years=1)

        now = utc_datetime_now()
        triad_year = (now - RELATIVE_YEAR).year if now.month < 3 else now.year
        this_month_start = Datetime(now.year, now.month, 1)
        last_month_start = this_month_start - relativedelta(months=1)
        last_month_finish = this_month_start - relativedelta(minutes=30)

        batch_reports = []
        config_contract = Contract.get_non_core_by_name(g.sess, "configuration")
        properties = config_contract.make_properties()
        if "g_supply_reports" in properties:
            for report_id in properties["g_supply_reports"]:
                batch_reports.append(Report.get_by_id(g.sess, report_id))

        note = truncated_line = None
        g_supply_note = (
            {"notes": []} if len(g_supply.note.strip()) == 0 else loads(g_supply.note)
        )

        notes = g_supply_note["notes"]
        if len(notes) > 0:
            note = notes[-1]
            lines = note["body"].splitlines()
            if len(lines) > 0:
                line0 = lines[0]
                if len(lines) > 1 or len(line0) > 50:
                    truncated_line = line0[:50]
        return render_template(
            "supply.html",
            triad_year=triad_year,
            now=now,
            last_month_start=last_month_start,
            last_month_finish=last_month_finish,
            g_era_bundles=g_era_bundles,
            g_supply=g_supply,
            system_properties=properties,
            truncated_line=truncated_line,
            note=note,
            this_month_start=this_month_start,
            batch_reports=batch_reports,
            debug=debug,
        )

    except BadRequest as e:
        flash(e.description)
        g.sess.rollback()
        return render_template("supply.html")


@gas.route("/supplies/<int:g_supply_id>/edit")
def supply_edit_get(g_supply_id):
    g_supply = GSupply.get_by_id(g.sess, g_supply_id)
    g_eras = (
        g.sess.query(GEra)
        .filter(GEra.g_supply == g_supply)
        .order_by(GEra.start_date.desc())
    )
    g_exit_zones = g.sess.query(GExitZone).order_by(GExitZone.code).all()
    return render_template(
        "supply_edit.html",
        g_supply=g_supply,
        g_eras=g_eras,
        g_exit_zones=g_exit_zones,
    )


@gas.route("/supplies/<int:g_supply_id>/edit", methods=["DELETE"])
def supply_edit_delete(g_supply_id):
    g_supply = GSupply.get_by_id(g.sess, g_supply_id)
    try:
        g_supply.delete(g.sess)
        g.sess.commit()
        return hx_redirect("/supplies")
    except BadRequest as e:
        flash(e.description)
        g_eras = g.sess.scalars(
            select(GEra)
            .where(GEra.g_supply == g_supply)
            .order_by(GEra.start_date.desc())
        )
        g_exit_zones = g.sess.query(GExitZone).order_by(GExitZone.code).all()
        return make_response(
            render_template(
                "supply_edit.html",
                g_supply=g_supply,
                g_eras=g_eras,
                g_exit_zones=g_exit_zones,
            ),
            400,
        )


@gas.route("/supplies/<int:g_supply_id>/edit", methods=["POST"])
def supply_edit_post(g_supply_id):
    g_supply = GSupply.get_by_id(g.sess, g_supply_id)
    try:
        if "insert_g_era" in request.values:
            start_date = req_date("start")
            g_supply.insert_g_era_at(g.sess, start_date)
            g.sess.commit()
            return chellow_redirect(f"/supplies/{g_supply.id}", 303)
        else:
            mprn = req_str("mprn")
            name = req_str("name")
            g_exit_zone_id = req_int("g_exit_zone_id")
            g_exit_zone = GExitZone.get_by_id(g.sess, g_exit_zone_id)
            g_supply.update(mprn, name, g_exit_zone)
            g.sess.commit()
            return chellow_redirect(f"/supplies/{g_supply.id}", 303)
    except BadRequest as e:
        flash(e.description)
        g_eras = (
            g.sess.query(GEra)
            .filter(GEra.g_supply == g_supply)
            .order_by(GEra.start_date.desc())
        )
        g_exit_zones = g.sess.query(GExitZone).order_by(GExitZone.code).all()
        return make_response(
            render_template(
                "supply_edit.html",
                g_supply=g_supply,
                g_eras=g_eras,
                g_exit_zones=g_exit_zones,
            ),
            400,
        )


@gas.route("/batches")
def batches_get():
    g_contract_id = req_int("g_contract_id")
    g_contract = GContract.get_by_id(g.sess, g_contract_id)
    g_batches = g.sess.execute(
        select(
            GBatch,
            func.count(GBill.id),
            func.coalesce(func.sum(GBill.net), 0),
            func.coalesce(func.sum(GBill.vat), 0),
            func.coalesce(func.sum(GBill.gross), 0),
            func.coalesce(func.sum(GBill.kwh), 0),
        )
        .join(GBill, isouter=True)
        .where(GBatch.g_contract == g_contract)
        .group_by(GBatch.id)
        .order_by(GBatch.reference.desc())
    )
    return render_template("batches.html", g_contract=g_contract, g_batches=g_batches)


@gas.route("/supplier_contracts/<int:g_contract_id>/add_batch")
def batch_add_get(g_contract_id):
    g_contract = GContract.get_by_id(g.sess, g_contract_id)
    g_batches = g.sess.execute(
        select(GBatch)
        .where(GBatch.g_contract == g_contract)
        .order_by(GBatch.reference.desc())
    ).scalars()
    return render_template("batch_add.html", g_contract=g_contract, g_batches=g_batches)


@gas.route("/supplier_contracts/<int:g_contract_id>/add_batch", methods=["POST"])
def batch_add_post(g_contract_id):
    try:
        g_contract = GContract.get_by_id(g.sess, g_contract_id)
        reference = req_str("reference")
        description = req_str("description")

        g_batch = g_contract.insert_g_batch(g.sess, reference, description)
        g.sess.commit()
        return chellow_redirect(f"/batches/{g_batch.id}", 303)
    except BadRequest as e:
        flash(e.description)
        g.sess.rollback()
        g_batches = g.sess.execute(
            select(GBatch)
            .where(GBatch.g_contract == g_contract)
            .order_by(GBatch.reference.desc())
        )
        return make_response(
            render_template(
                "batch_add.html", g_contract=g_contract, g_batches=g_batches
            ),
            400,
        )


@gas.route("/batches/<int:g_batch_id>")
def batch_get(g_batch_id):
    g_batch = GBatch.get_by_id(g.sess, g_batch_id)

    num_bills = sum_net_gbp = sum_vat_gbp = sum_gross_gbp = sum_kwh = 0
    vat_breakdown = {}
    max_reads = 0

    g_bills = (
        g.sess.execute(
            select(GBill)
            .where(GBill.g_batch == g_batch)
            .order_by(GBill.reference, GBill.start_date)
        )
        .scalars()
        .all()
    )

    for g_bill in g_bills:
        num_bills += 1
        sum_net_gbp += g_bill.net
        sum_vat_gbp += g_bill.vat
        sum_gross_gbp += g_bill.gross
        sum_kwh += g_bill.kwh
        max_reads = max(len(g_bill.g_reads), max_reads)

        if g_bill.vat != 0:
            bd = g_bill.make_breakdown()

            if "vat_rate" in bd:
                vat_tuple = tuple([str(v * 100) for v in sorted(bd["vat_rate"])])
                vat_percentage = vat_tuple[0] if len(vat_tuple) == 1 else vat_tuple
                try:
                    vbd = vat_breakdown[vat_percentage]
                except KeyError:
                    vbd = vat_breakdown[vat_percentage] = defaultdict(int)

                vbd["vat"] += g_bill.vat
                vbd["net"] += g_bill.net

            if "vat" in bd:
                for vat_percentage, vat_bd in bd["vat"].items():
                    try:
                        vbd = vat_breakdown[vat_percentage]
                    except KeyError:
                        vbd = vat_breakdown[vat_percentage] = defaultdict(int)

                    vbd["vat"] += vat_bd.get("vat", Decimal("0"))
                    vbd["net"] += vat_bd.get("net", Decimal("0"))

    config_contract = Contract.get_non_core_by_name(g.sess, "configuration")
    properties = config_contract.make_properties()

    if "g_batch_reports" in properties:
        g_batch_reports = []
        for report_id in properties["g_batch_reports"]:
            g_batch_reports.append(Report.get_by_id(g.sess, report_id))
    else:
        g_batch_reports = None

    return render_template(
        "batch.html",
        g_batch_reports=g_batch_reports,
        g_batch=g_batch,
        g_bills=g_bills,
        max_reads=max_reads,
        num_bills=num_bills,
        sum_net_gbp=sum_net_gbp,
        sum_vat_gbp=sum_vat_gbp,
        sum_gross_gbp=sum_gross_gbp,
        sum_kwh=sum_kwh,
        vat_breakdown=vat_breakdown,
    )


@gas.route("/batches/<int:g_batch_id>/edit")
def batch_edit_get(g_batch_id):
    g_batch = GBatch.get_by_id(g.sess, g_batch_id)
    return render_template("batch_edit.html", g_batch=g_batch)


@gas.route("/batches/<int:g_batch_id>/edit", methods=["POST"])
def batch_edit_post(g_batch_id):
    try:
        g_batch = GBatch.get_by_id(g.sess, g_batch_id)
        if "update" in request.values:
            reference = req_str("reference")
            description = req_str("description")
            g_batch.update(g.sess, reference, description)
            g.sess.commit()
            return chellow_redirect(f"/batches/{g_batch.id}", 303)
        elif "delete_bills" in request.values:
            g.sess.query(GBill).filter(GBill.g_batch == g_batch).delete(False)
            g.sess.commit()
            return hx_redirect(f"/batches/{g_batch.id}")
    except BadRequest as e:
        flash(e.description)
        return make_response(render_template("batch_edit.html", g_batch=g_batch), 400)


@gas.route("/batches/<int:g_batch_id>/edit", methods=["DELETE"])
def batch_edit_delete(g_batch_id):
    try:
        g_batch = GBatch.get_by_id(g.sess, g_batch_id)
        g_contract = g_batch.g_contract
        g_batch.delete(g.sess)
        g.sess.commit()
        return hx_redirect(f"/batches?g_contract_id={g_contract.id}")
    except BadRequest as e:
        flash(e.description)
        return make_response(render_template("batch_edit.html", g_batch=g_batch), 400)


@gas.route("/bills/<int:g_bill_id>")
def bill_get(g_bill_id):
    g_bill = GBill.get_by_id(g.sess, g_bill_id)
    g_reads = (
        g.sess.query(GRegisterRead)
        .filter(GRegisterRead.g_bill == g_bill)
        .order_by(GRegisterRead.pres_date.desc())
    )
    fields = {"g_bill": g_bill, "g_reads": g_reads}
    try:
        breakdown = g_bill.make_breakdown()

        raw_lines = g_bill.raw_lines

        rows = set()
        columns = set()
        grid = defaultdict(dict)

        for k, v in tuple(breakdown.items()):
            if k.endswith("_gbp"):
                columns.add("gbp")
                row_name = k[:-4]
                rows.add(row_name)
                grid[row_name]["gbp"] = v
                del breakdown[k]

        for k, v in tuple(breakdown.items()):
            for row_name in sorted(list(rows), key=len, reverse=True):
                if k.startswith(row_name + "_"):
                    col_name = k[len(row_name) + 1 :]
                    columns.add(col_name)
                    grid[row_name][col_name] = csv_make_val(v)
                    del breakdown[k]
                    break

        for k, v in breakdown.items():
            pair = k.split("_")
            row_name = "_".join(pair[:-1])
            column_name = pair[-1]
            rows.add(row_name)
            columns.add(column_name)
            grid[row_name][column_name] = csv_make_val(v)

        column_list = sorted(list(columns))
        for rate_name in [col for col in column_list if col.endswith("rate")]:
            column_list.remove(rate_name)
            column_list.append(rate_name)

        if "gbp" in column_list:
            column_list.remove("gbp")
            column_list.append("gbp")

        row_list = sorted(list(rows))
        fields.update(
            {
                "raw_lines": raw_lines,
                "row_list": row_list,
                "column_list": column_list,
                "grid": grid,
            }
        )
    except SyntaxError:
        pass

    return render_template("bill.html", **fields)


@gas.route("/bill_imports")
def bill_imports_get():
    g_batch_id = req_int("g_batch_id")
    g_batch = GBatch.get_by_id(g.sess, g_batch_id)
    importer_ids = sorted(
        chellow.gas.bill_import.get_bill_importer_ids(g_batch.id), reverse=True
    )
    parser_names = chellow.gas.bill_import.find_parser_names()
    g_contract_props = g_batch.g_contract.make_properties()

    return render_template(
        "bill_imports.html",
        importer_ids=importer_ids,
        g_batch=g_batch,
        parser_names=parser_names,
        default_bill_parser_name=g_contract_props.get("default_bill_parser", ""),
    )


@gas.route("/bill_imports", methods=["POST"])
def bill_imports_post():
    try:
        g_batch_id = req_int("g_batch_id")
        g_batch = GBatch.get_by_id(g.sess, g_batch_id)
        file_item = req_file("import_file")
        file_bytes = file_item.stream.read()
        parser_name = req_str("parser_name")
        imp_id = chellow.gas.bill_import.start_bill_importer(
            g.sess, g_batch.id, file_item.filename, file_bytes, parser_name
        )
        return chellow_redirect(f"/bill_imports/{imp_id}", 303)
    except BadRequest as e:
        flash(e.description)
        importer_ids = sorted(
            chellow.gas.bill_import.get_bill_importer_ids(g_batch.id), reverse=True
        )
        g_contract_props = g_batch.g_contract.make_properties()
        return make_response(
            render_template(
                "bill_imports.html",
                importer_ids=importer_ids,
                g_batch=g_batch,
                parser_names=chellow.gas.bill_import.find_parser_names(),
                default_bill_parser_name=g_contract_props.get(
                    "default_bill_parser", ""
                ),
            ),
            400,
        )


@gas.route("/bill_imports/<int:imp_id>")
def bill_import_get(imp_id):
    importer = chellow.gas.bill_import.get_bill_importer(imp_id)
    g_batch = GBatch.get_by_id(g.sess, importer.g_batch_id)
    fields = {"g_batch": g_batch}
    if importer is not None:
        imp_fields = importer.make_fields()
        if "successful_bills" in imp_fields and len(imp_fields["successful_bills"]) > 0:
            fields["successful_max_registers"] = max(
                len(bill["reads"]) for bill in imp_fields["successful_bills"]
            )
        fields.update(imp_fields)
        fields["status"] = importer.status()
    return render_template("bill_import.html", **fields)


@gas.route("/batches/<int:g_batch_id>/add_bill")
def bill_add_get(g_batch_id):
    g_batch = GBatch.get_by_id(g.sess, g_batch_id)
    g_bills = (
        g.sess.query(GBill).filter(GBill.g_batch == g_batch).order_by(GBill.start_date)
    )
    bill_types = g.sess.query(BillType).order_by(BillType.code).all()
    return render_template(
        "bill_add.html", g_batch=g_batch, g_bills=g_bills, bill_types=bill_types
    )


@gas.route("/batches/<int:g_batch_id>/add_bill", methods=["POST"])
def bill_add_post(g_batch_id):
    try:
        g_batch = GBatch.get_by_id(g.sess, g_batch_id)
        bill_type_id = req_int("bill_type_id")
        bill_type = BillType.get_by_id(g.sess, bill_type_id)
        mprn = req_str("mprn")
        g_supply = GSupply.get_by_mprn(g.sess, mprn)
        account = req_str("account")
        reference = req_str("reference")
        issue_date = req_date("issue")
        start_date = req_date("start")
        finish_date = req_date("finish")
        kwh = req_decimal("kwh")
        net = req_decimal("net")
        vat = req_decimal("vat")
        gross = req_decimal("gross")
        breakdown = req_zish("breakdown")
        g_bill = g_batch.insert_g_bill(
            g.sess,
            g_supply,
            bill_type,
            reference,
            account,
            issue_date,
            start_date,
            finish_date,
            kwh,
            net,
            vat,
            gross,
            "",
            breakdown,
        )
        g.sess.commit()
        return chellow_redirect(f"/bills/{g_bill.id}", 303)
    except BadRequest as e:
        flash(e.description)
        g.sess.rollback()
        g_bills = (
            g.sess.query(GBill)
            .filter(GBill.g_batch == g_batch)
            .order_by(GBill.start_date)
        )
        bill_types = g.sess.query(BillType).order_by(BillType.code).all()
        return make_response(
            render_template(
                "bill_add.html",
                g_batch=g_batch,
                g_bills=g_bills,
                bill_types=bill_types,
            ),
            400,
        )


@gas.route("/supplies/<int:g_supply_id>/notes")
def supply_notes_get(g_supply_id):
    g_supply = GSupply.get_by_id(g.sess, g_supply_id)

    if len(g_supply.note.strip()) > 0:
        g_supply_note = loads(g_supply.note)
    else:
        g_supply_note = {"notes": []}

    return render_template(
        "supply_notes.html", g_supply=g_supply, g_supply_note=g_supply_note
    )


@gas.route("/supplies/<int:g_supply_id>/notes/add")
def supply_note_add_get(g_supply_id):
    g_supply = GSupply.get_by_id(g.sess, g_supply_id)
    return render_template("supply_note_add.html", g_supply=g_supply)


@gas.route("/supplies/<int:g_supply_id>/notes/add", methods=["POST"])
def supply_note_add_post(g_supply_id):
    try:
        g_supply = GSupply.get_by_id(g.sess, g_supply_id)
        body = req_str("body")
        if len(g_supply.note.strip()) == 0:
            note_dict = {"notes": []}
        else:
            note_dict = loads(g_supply.note)

        note_dict["notes"].append({"body": body, "timestamp": utc_datetime_now()})
        g_supply.note = dumps(note_dict)
        g.sess.commit()
        return chellow_redirect(f"/supplies/{g_supply_id}", 303)
    except BadRequest as e:
        flash(e.description)
        return make_response(
            render_template("supply_note_add.html", g_supply=g_supply), 400
        )


@gas.route("/supplies/<int:g_supply_id>/notes/<int:index>/edit")
def supply_note_edit_get(g_supply_id, index):
    g_supply = GSupply.get_by_id(g.sess, g_supply_id)
    g_supply_note = loads(g_supply.note)
    note = g_supply_note["notes"][index]
    note["index"] = index
    return render_template("supply_note_edit.html", g_supply=g_supply, note=note)


@gas.route("/supplies/<int:g_supply_id>/notes/<int:index>/edit", methods=["DELETE"])
def supply_note_edit_delete(g_supply_id, index):
    try:
        g_supply = GSupply.get_by_id(g.sess, g_supply_id)
        g_supply_note = loads(g_supply.note)
        del g_supply_note["notes"][index]
        g_supply.note = dumps(g_supply_note)
        g.sess.commit()
        return hx_redirect(f"/supplies/{g_supply_id}/notes")
    except BadRequest as e:
        flash(e.description)
        g_supply_note = loads(g_supply.note)
        note = g_supply_note["notes"][index]
        note["index"] = index
        return render_template("supply_note_edit.html", g_supply=g_supply, note=note)


@gas.route("/supplies/<int:g_supply_id>/notes/<int:index>/edit", methods=["POST"])
def supply_note_edit_post(g_supply_id, index):
    try:
        g_supply = GSupply.get_by_id(g.sess, g_supply_id)
        g_supply_note = loads(g_supply.note)
        body = req_str("body")
        note = g_supply_note["notes"][index]
        note["body"] = body
        g_supply.note = dumps(g_supply_note)
        g.sess.commit()
        return hx_redirect(f"/supplies/{g_supply_id}/notes/{index}")
    except BadRequest as e:
        flash(e.description)
        g_supply_note = loads(g_supply.note)
        note = g_supply_note["notes"][index]
        note["index"] = index
        return render_template("supply_note_edit.html", g_supply=g_supply, note=note)


@gas.route("/eras/<int:g_era_id>/edit")
def era_edit_get(g_era_id):
    g_era = GEra.get_by_id(g.sess, g_era_id)
    supplier_g_contracts = g.sess.scalars(
        select(GContract)
        .where(GContract.is_industry == false())
        .order_by(GContract.name)
    )
    site_g_eras = (
        g.sess.query(SiteGEra)
        .join(Site)
        .filter(SiteGEra.g_era == g_era)
        .order_by(Site.code)
        .all()
    )
    g_units = g.sess.query(GUnit).order_by(GUnit.code).all()
    g_reading_frequencies = (
        g.sess.query(GReadingFrequency).order_by(GReadingFrequency.code).all()
    )
    return render_template(
        "era_edit.html",
        g_era=g_era,
        supplier_g_contracts=supplier_g_contracts,
        site_g_eras=site_g_eras,
        g_units=g_units,
        g_reading_frequencies=g_reading_frequencies,
    )


@gas.route("/eras/<int:g_era_id>/edit", methods=["POST"])
def era_edit_post(g_era_id):
    try:
        g_era = GEra.get_by_id(g.sess, g_era_id)

        if "delete" in request.values:
            g_supply = g_era.g_supply
            g_supply.delete_g_era(g.sess, g_era)
            g.sess.commit()
            return chellow_redirect(f"/supplies/{g_supply.id}", 303)

        elif "attach" in request.values:
            site_code = req_str("site_code")
            site = Site.get_by_code(g.sess, site_code)
            g_era.attach_site(g.sess, site)
            g.sess.commit()
            return chellow_redirect(f"/supplies/{g_era.g_supply.id}", 303)

        elif "detach" in request.values:
            site_id = req_int("site_id")
            site = Site.get_by_id(g.sess, site_id)
            g_era.detach_site(g.sess, site)
            g.sess.commit()
            return chellow_redirect(f"/supplies/{g_era.g_supply.id}", 303)

        elif "locate" in request.values:
            site_id = req_int("site_id")
            site = Site.get_by_id(g.sess, site_id)
            g_era.set_physical_location(g.sess, site)
            g.sess.commit()
            return chellow_redirect(f"/supplies/{g_era.g_supply.id}", 303)

        else:
            start_date = req_date("start")
            is_ended = req_checkbox("is_ended")
            finish_date = req_date("finish") if is_ended else None
            msn = req_str("msn")
            correction_factor = req_decimal("correction_factor")
            g_contract_id = req_int("g_contract_id")
            g_contract = GContract.get_by_id(g.sess, g_contract_id)
            account = req_str("account")
            g_unit_id = req_int("g_unit_id")
            g_unit = GUnit.get_by_id(g.sess, g_unit_id)
            g_reading_frequency_id = req_int("g_reading_frequency_id")
            g_reading_frequency = GReadingFrequency.get_by_id(
                g.sess, g_reading_frequency_id
            )
            aq = req_decimal("aq")
            soq = req_decimal("soq")
            g_era.g_supply.update_g_era(
                g.sess,
                g_era,
                start_date,
                finish_date,
                msn,
                correction_factor,
                g_unit,
                g_contract,
                account,
                g_reading_frequency,
                aq,
                soq,
            )
            g.sess.commit()
            return chellow_redirect(f"/supplies/{g_era.g_supply.id}", 303)

    except BadRequest as e:
        flash(e.description)
        supplier_g_contracts = g.sess.query(GContract).order_by(GContract.name)
        site_g_eras = (
            g.sess.query(SiteGEra)
            .join(Site)
            .filter(SiteGEra.g_era == g_era)
            .order_by(Site.code)
            .all()
        )
        g_units = g.sess.query(GUnit).order_by(GUnit.code).all()
        g_reading_frequencies = (
            g.sess.query(GReadingFrequency).order_by(GReadingFrequency.code).all()
        )
        return make_response(
            render_template(
                "era_edit.html",
                g_era=g_era,
                supplier_g_contracts=supplier_g_contracts,
                site_g_eras=site_g_eras,
                g_units=g_units,
                g_reading_frequencies=g_reading_frequencies,
            ),
            400,
        )


@gas.route("/bills/<int:g_bill_id>/edit")
def bill_edit_get(g_bill_id):
    g_bill = GBill.get_by_id(g.sess, g_bill_id)
    bill_types = g.sess.query(BillType).order_by(BillType.code).all()
    return render_template("bill_edit.html", g_bill=g_bill, bill_types=bill_types)


@gas.route("/bills/<int:g_bill_id>/edit", methods=["POST"])
def bill_edit_post(g_bill_id):
    try:
        g_bill = GBill.get_by_id(g.sess, g_bill_id)
        if "delete" in request.values:
            g_batch = g_bill.g_batch
            g_bill.delete(g.sess)
            g.sess.commit()
            return chellow_redirect(f"/batches/{g_batch.id}", 303)

        else:
            account = req_str("account")
            reference = req_str("reference")
            issue_date = req_date("issue")
            start_date = req_date("start")
            finish_date = req_date("finish")
            kwh = req_decimal("kwh")
            net_gbp = req_decimal("net_gbp")
            vat_gbp = req_decimal("vat_gbp")
            gross_gbp = req_decimal("gross_gbp")
            type_id = req_int("bill_type_id")
            raw_lines = req_str("raw_lines")
            breakdown = req_zish("breakdown")
            bill_type = BillType.get_by_id(g.sess, type_id)

            g_bill.update(
                bill_type,
                reference,
                account,
                issue_date,
                start_date,
                finish_date,
                kwh,
                net_gbp,
                vat_gbp,
                gross_gbp,
                raw_lines,
                breakdown,
            )
            g.sess.commit()
            return chellow_redirect(f"/bills/{g_bill.id}", 303)

    except BadRequest as e:
        flash(e.description)
        g_bill = GBill.get_by_id(g.sess, g_bill_id)
        bill_types = g.sess.query(BillType).order_by(BillType.code).all()
        return make_response(
            render_template("bill_edit.html", g_bill=g_bill, bill_types=bill_types),
            400,
        )


@gas.route("/bills/<int:g_bill_id>/add_read")
def read_add_get(g_bill_id):
    g_read_types = g.sess.query(GReadType).order_by(GReadType.code)
    g_bill = GBill.get_by_id(g.sess, g_bill_id)
    g_units = g.sess.query(GUnit).order_by(GUnit.code)
    return render_template(
        "read_add.html", g_bill=g_bill, g_read_types=g_read_types, g_units=g_units
    )


@gas.route("/bills/<int:g_bill_id>/add_read", methods=["POST"])
def g_read_add_post(g_bill_id):
    try:
        g_bill = GBill.get_by_id(g.sess, g_bill_id)
        msn = req_str("msn")
        g_unit_id = req_int("g_unit_id")
        g_unit = GUnit.get_by_id(g.sess, g_unit_id)
        correction_factor = req_decimal("correction_factor")
        calorific_value = req_decimal("calorific_value")
        prev_date = req_date("prev_date")
        prev_value = req_decimal("prev_value")
        prev_type_id = req_int("prev_type_id")
        prev_type = GReadType.get_by_id(g.sess, prev_type_id)
        pres_date = req_date("pres_date")
        pres_value = req_decimal("pres_value")
        pres_type_id = req_int("pres_type_id")
        pres_type = GReadType.get_by_id(g.sess, pres_type_id)

        g_bill.insert_g_read(
            g.sess,
            msn,
            g_unit,
            correction_factor,
            calorific_value,
            prev_value,
            prev_date,
            prev_type,
            pres_value,
            pres_date,
            pres_type,
        )
        g.sess.commit()
        return chellow_redirect(f"/bills/{g_bill.id}", 303)
    except BadRequest as e:
        flash(e.description)
        g_read_types = g.sess.query(GReadType).order_by(GReadType.code)
        return render_template(
            "read_add.html", g_bill=g_bill, g_read_types=g_read_types
        )


@gas.route("/reads/<int:g_read_id>/edit")
def read_edit_get(g_read_id):
    g_read_types = g.sess.query(GReadType).order_by(GReadType.code).all()
    g_read = GRegisterRead.get_by_id(g.sess, g_read_id)
    g_units = g.sess.query(GUnit).order_by(GUnit.code)
    return render_template(
        "read_edit.html", g_read=g_read, g_read_types=g_read_types, g_units=g_units
    )


@gas.route("/reads/<int:g_read_id>/edit", methods=["POST"])
def read_edit_post(g_read_id):
    try:
        g_read = GRegisterRead.get_by_id(g.sess, g_read_id)
        if "update" in request.values:
            msn = req_str("msn")
            prev_date = req_date("prev_date")
            prev_value = req_decimal("prev_value")
            prev_type_id = req_int("prev_type_id")
            prev_type = GReadType.get_by_id(g.sess, prev_type_id)
            pres_date = req_date("pres_date")
            pres_value = req_decimal("pres_value")
            pres_type_id = req_int("pres_type_id")
            pres_type = GReadType.get_by_id(g.sess, pres_type_id)
            g_unit_id = req_int("g_unit_id")
            g_unit = GUnit.get_by_id(g.sess, g_unit_id)
            correction_factor = req_decimal("correction_factor")
            calorific_value = req_decimal("calorific_value")

            g_read.update(
                msn,
                g_unit,
                correction_factor,
                calorific_value,
                prev_value,
                prev_date,
                prev_type,
                pres_value,
                pres_date,
                pres_type,
            )

            g.sess.commit()
            return chellow_redirect(f"/bills/{g_read.g_bill.id}", 303)

        elif "delete" in request.values:
            g_bill = g_read.g_bill
            g_read.delete(g.sess)
            g.sess.commit()
            return chellow_redirect(f"/bills/{g_bill.id}", 303)

    except BadRequest as e:
        flash(e.description)
        g_read_types = g.sess.query(GReadType).order_by(GReadType.code).all()
        g_units = g.sess.query(GUnit).order_by(GUnit.code)
        return make_response(
            render_template(
                "read_edit.html",
                g_read=g_read,
                g_read_types=g_read_types,
                g_units=g_units,
            ),
            400,
        )


@gas.route("/units")
def units_get():
    g_units = g.sess.query(GUnit).order_by(GUnit.code)
    return render_template("units.html", g_units=g_units)


@gas.route("/units/<int:g_unit_id>")
def unit_get(g_unit_id):
    g_unit = GUnit.get_by_id(g.sess, g_unit_id)
    return render_template("unit.html", g_unit=g_unit)


@gas.route("/read_types/<int:g_read_type_id>")
def read_type_get(g_read_type_id):
    g_read_type = GReadType.get_by_id(g.sess, g_read_type_id)
    return render_template("read_type.html", g_read_type=g_read_type)


@gas.route("/read_types")
def read_types_get():
    g_read_types = g.sess.query(GReadType).order_by(GReadType.code)
    return render_template("read_types.html", g_read_types=g_read_types)


@gas.route("/reports")
def reports_get():
    now = utc_datetime_now()
    now_day = utc_datetime(now.year, now.month, now.day)
    month_start = Datetime(now.year, now.month, 1) - relativedelta(months=1)
    month_finish = Datetime(now.year, now.month, 1) - HH
    return render_template(
        "reports.html",
        month_start=month_start,
        month_finish=month_finish,
        now_day=now_day,
    )


@gas.route("/dns")
def dns_get():
    g_dns = g.sess.query(GDn).order_by(GDn.code)
    return render_template("dns.html", g_dns=g_dns)


@gas.route("/dns/<int:g_dn_id>")
def dn_get(g_dn_id):
    g_dn = GDn.get_by_id(g.sess, g_dn_id)
    g_ldzs = g.sess.query(GLdz).filter(GLdz.g_dn == g_dn).order_by(GLdz.code).all()
    return render_template("dn.html", g_dn=g_dn, g_ldzs=g_ldzs)


@gas.route("/industry_contracts")
def industry_contracts_get():
    contracts = g.sess.execute(
        select(GContract)
        .where(GContract.is_industry == true())
        .order_by(GContract.name)
    ).scalars()
    return render_template("industry_contracts.html", contracts=contracts)


@gas.route("/industry_contracts/<int:contract_id>")
def industry_contract_get(contract_id):
    contract = GContract.get_industry_by_id(g.sess, contract_id)
    rate_scripts = g.sess.execute(
        select(GRateScript)
        .where(GRateScript.g_contract == contract)
        .order_by(GRateScript.start_date.desc())
    ).scalars()

    try:
        import_module(f"chellow.gas.{contract.name}").get_importer()
        has_auto_importer = True
    except AttributeError:
        has_auto_importer = False
    except ImportError:
        has_auto_importer = False

    return render_template(
        "industry_contract.html",
        contract=contract,
        rate_scripts=rate_scripts,
        has_auto_importer=has_auto_importer,
    )


@gas.route("/industry_contracts/add")
def industry_contract_add_get():
    contracts = g.sess.execute(
        select(GContract)
        .where(GContract.is_industry == true())
        .order_by(GContract.name)
    )
    return render_template("industry_contract_add.html", contracts=contracts)


@gas.route("/industry_contracts/add", methods=["POST"])
def industry_contract_add_post():
    try:
        name = req_str("name")
        start_date = req_date("start")
        charge_script = req_str("charge_script")
        properties = req_zish("properties")

        contract = GContract.insert(
            g.sess, True, name, charge_script, properties, start_date, None, {}
        )
        g.sess.commit()
        return chellow_redirect(f"/industry_contracts/{contract.id}", 303)

    except BadRequest as e:
        flash(e.description)
        g.sess.rollback()
        contracts = g.sess.query(GContract).order_by(GContract.name)
        return make_response(
            render_template("industry_contract_add.html", contracts=contracts), 400
        )


@gas.route("/industry_contracts/<int:g_contract_id>/edit")
def industry_contract_edit_get(g_contract_id):
    g_contract = GContract.get_industry_by_id(g.sess, g_contract_id)
    return render_template("industry_contract_edit.html", g_contract=g_contract)


@gas.route("/industry_contracts/<int:g_contract_id>/edit", methods=["DELETE"])
def industry_contract_edit_delete(g_contract_id):
    try:
        g_contract = GContract.get_industry_by_id(g.sess, g_contract_id)
        g_contract.delete(g.sess)
        g.sess.commit()
        return hx_redirect("/industry_contracts")
    except BadRequest as e:
        flash(e.description)
        g.sess.rollback()
        return make_response(
            render_template("industry_contract_edit.html", g_contract=g_contract), 400
        )


@gas.route("/industry_contracts/<int:g_contract_id>/edit", methods=["PATCH"])
def industry_contract_edit_patch(g_contract_id):
    try:
        g_contract = GContract.get_industry_by_id(g.sess, g_contract_id)
        if "state" in request.values:
            state = req_zish("state")
            g_contract.update_state(state)
        else:
            name = req_str("name")
            charge_script = req_str("charge_script")
            properties = req_zish("properties")
            g_contract.update(name, charge_script, properties)

        g.sess.commit()
        return hx_redirect(f"/industry_contracts/{g_contract.id}", 303)
    except BadRequest as e:
        flash(e.description)
        g.sess.rollback()
        return make_response(
            render_template("industry_contract_edit.html", g_contract=g_contract), 400
        )


@gas.route("/industry_contracts/<int:contract_id>/auto_importer")
def industry_auto_importer_get(contract_id):
    contract = GContract.get_industry_by_id(g.sess, contract_id)
    importer = import_module(f"chellow.gas.{contract.name}").get_importer()
    return render_template(
        "industry_auto_importer.html", importer=importer, contract=contract
    )


@gas.route("/industry_contracts/<int:contract_id>/auto_importer", methods=["POST"])
def industry_auto_importer_post(contract_id):
    try:
        contract = GContract.get_industry_by_id(g.sess, contract_id)
        importer = import_module(f"chellow.gas.{contract.name}").get_importer()
        importer.go()
        return chellow_redirect(f"/industry_contracts/{contract.id}/auto_importer", 303)
    except BadRequest as e:
        g.sess.rollback()
        flash(e.description)
        return make_response(
            render_template(
                "industry_auto_importer.html", importer=importer, contract=contract
            ),
            400,
        )


@gas.route("/industry_contracts/<int:g_contract_id>/add_rate_script")
def industry_rate_script_add_get(g_contract_id):
    now = utc_datetime_now()
    initial_date = utc_datetime(now.year, now.month)
    g_contract = GContract.get_industry_by_id(g.sess, g_contract_id)
    return render_template(
        "industry_rate_script_add.html",
        g_contract=g_contract,
        initial_date=initial_date,
    )


@gas.route("/industry_contracts/<int:g_contract_id>/add_rate_script", methods=["POST"])
def industry_rate_script_add_post(g_contract_id):
    try:
        g_contract = GContract.get_by_id(g.sess, g_contract_id)
        start_date = req_date("start")
        g_rate_script = g_contract.insert_g_rate_script(g.sess, start_date, {})
        g.sess.commit()
        return chellow_redirect(f"/industry_rate_scripts/{g_rate_script.id}", 303)
    except BadRequest as e:
        flash(e.description)
        now = utc_datetime_now()
        initial_date = utc_datetime(now.year, now.month)
        return make_response(
            render_template(
                "industry_rate_script_add.html",
                g_contract=g_contract,
                initial_date=initial_date,
            ),
            400,
        )


@gas.route("/industry_rate_scripts/<int:g_rate_script_id>/edit")
def industry_rate_script_edit_get(g_rate_script_id):
    g_rate_script = GRateScript.get_by_id(g.sess, g_rate_script_id)
    return render_template(
        "industry_rate_script_edit.html", g_rate_script=g_rate_script
    )


@gas.route("/industry_rate_scripts/<int:g_rate_script_id>/edit", methods=["POST"])
def industry_rate_script_edit_post(g_rate_script_id):
    try:
        g_rate_script = GRateScript.get_by_id(g.sess, g_rate_script_id)
        g_contract = g_rate_script.g_contract
        if "delete" in request.values:
            g_contract.delete_g_rate_script(g.sess, g_rate_script)
            g.sess.commit()
            return chellow_redirect(f"/industry_contracts/{g_contract.id}", 303)
        elif "import" in request.form:
            file_item = req_file("dn_file")

            rates = chellow.gas.dn_rate_parser.find_rates(
                file_item.filename, BytesIO(file_item.read())
            )
            g_contract.update_g_rate_script(
                g.sess,
                g_rate_script,
                g_rate_script.start_date,
                g_rate_script.finish_date,
                rates,
            )
            g.sess.commit()
            return chellow_redirect(f"/industry_rate_scripts/{g_rate_script.id}", 303)
        else:
            script = req_zish("script")
            start_date = req_date("start")
            has_finished = req_checkbox("has_finished")
            finish_date = req_date("finish") if has_finished else None
            g_contract.update_g_rate_script(
                g.sess, g_rate_script, start_date, finish_date, script
            )
            g.sess.commit()
            return chellow_redirect(f"/industry_rate_scripts/{g_rate_script.id}", 303)
    except BadRequest as e:
        flash(e.description)
        g.sess.rollback()
        return make_response(
            render_template(
                "industry_rate_script_edit.html", g_rate_script=g_rate_script
            ),
            400,
        )


@gas.route("/industry_rate_scripts/<int:g_rate_script_id>")
def industry_rate_script_get(g_rate_script_id):
    g_rate_script = GRateScript.get_by_id(g.sess, g_rate_script_id)
    g_contract = g_rate_script.g_contract
    next_g_rate_script = (
        g.sess.execute(
            select(GRateScript)
            .where(
                GRateScript.g_contract == g_contract,
                GRateScript.start_date > g_rate_script.start_date,
            )
            .order_by(GRateScript.start_date)
        )
        .scalars()
        .first()
    )
    previous_g_rate_script = (
        g.sess.execute(
            select(GRateScript)
            .where(
                GRateScript.g_contract == g_contract,
                GRateScript.start_date < g_rate_script.start_date,
            )
            .order_by(GRateScript.start_date.desc())
        )
        .scalars()
        .first()
    )
    return render_template(
        "industry_rate_script.html",
        g_rate_script=g_rate_script,
        previous_g_rate_script=previous_g_rate_script,
        next_g_rate_script=next_g_rate_script,
    )


@gas.route("/ldzs/<int:g_ldz_id>")
def ldz_get(g_ldz_id):
    g_ldz = GLdz.get_by_id(g.sess, g_ldz_id)
    g_exit_zones = (
        g.sess.query(GExitZone)
        .filter(GExitZone.g_ldz == g_ldz)
        .order_by(GExitZone.code)
        .all()
    )
    return render_template("ldz.html", g_ldz=g_ldz, g_exit_zones=g_exit_zones)


@gas.route("/reading_frequencies/<int:g_reading_frequency_id>")
def reading_frequency_get(g_reading_frequency_id):
    g_reading_frequency = GReadingFrequency.get_by_id(g.sess, g_reading_frequency_id)
    return render_template(
        "reading_frequency.html", g_reading_frequency=g_reading_frequency
    )


@gas.route("/reading_frequencies")
def reading_frequencies_get():
    g_reading_frequencies = g.sess.query(GReadingFrequency).order_by(
        GReadingFrequency.code
    )
    return render_template(
        "reading_frequencies.html", g_reading_frequencies=g_reading_frequencies
    )


@gas.route("/exit_zones/<int:g_exit_zone_id>")
def exit_zone_get(g_exit_zone_id):
    g_exit_zone = GExitZone.get_by_id(g.sess, g_exit_zone_id)
    return render_template("exit_zone.html", g_exit_zone=g_exit_zone)


@gas.route("/batches/<int:g_batch_id>/csv")
def batch_csv_get(g_batch_id):
    g_batch = GBatch.get_by_id(g.sess, g_batch_id)
    rows = []
    max_reads = 0
    for g_bill in g.sess.scalars(
        select(GBill)
        .where(GBill.g_batch == g_batch)
        .order_by(GBill.reference, GBill.start_date)
        .options(
            joinedload(GBill.bill_type),
            joinedload(GBill.g_reads).joinedload(GRegisterRead.g_unit),
            joinedload(GBill.g_reads).joinedload(GRegisterRead.pres_type),
            joinedload(GBill.g_reads).joinedload(GRegisterRead.prev_type),
        )
    ).unique():
        row = [
            g_batch.g_contract.name,
            g_batch.reference,
            g_bill.reference,
            g_bill.account,
            g_bill.issue_date,
            g_bill.start_date,
            g_bill.finish_date,
            g_bill.kwh,
            g_bill.net,
            g_bill.vat,
            g_bill.gross,
            g_bill.bill_type.code,
            g_bill.breakdown,
        ]
        g_reads = g_bill.g_reads
        max_reads = max(max_reads, len(g_reads))
        for g_read in g_reads:
            row.extend(
                [
                    g_read.msn,
                    g_read.g_unit.code,
                    g_read.correction_factor,
                    g_read.calorific_value,
                    g_read.prev_date,
                    g_read.prev_value,
                    g_read.prev_type.code,
                    g_read.pres_date,
                    g_read.pres_value,
                    g_read.pres_type.code,
                ]
            )
        rows.append(row)

    si = StringIO()
    cw = csv.writer(si)
    read_titles = [
        "msn",
        "unit",
        "correction_factor",
        "calorific_value",
        "prev_date",
        "prev_value",
        "prev_type",
        "pres_date",
        "pres_value",
        "pres_type",
    ]
    titles = [
        "Contract",
        "Batch Reference",
        "Bill Reference",
        "Account",
        "Issued",
        "From",
        "To",
        "kWh",
        "Net",
        "VAT",
        "Gross",
        "Type",
        "breakdown",
    ]
    for i in range(max_reads):
        for rt in read_titles:
            titles.append(f"{i}_{rt}")
    cw.writerow(titles)
    for row in rows:
        cw.writerow(csv_make_val(v) for v in row)
    disp = f'attachment; filename="g_batch_{g_batch.id}.csv"'
    output = make_response(si.getvalue())
    output.headers["Content-Disposition"] = disp
    output.headers["Content-type"] = "text/csv"
    return output


@gas.route("/supplier_contracts")
def supplier_contracts_get():
    GRateScriptAliasFinish = aliased(GRateScript)

    current_contracts = g.sess.scalars(
        select(GContract)
        .join(
            GRateScriptAliasFinish,
            GContract.finish_g_rate_script_id == GRateScriptAliasFinish.id,
        )
        .where(
            GContract.is_industry == false(),
            GRateScriptAliasFinish.finish_date == null(),
        )
        .order_by(GContract.name)
    )
    ended_contracts = g.sess.scalars(
        select(GContract)
        .join(
            GRateScriptAliasFinish,
            GContract.finish_g_rate_script_id == GRateScriptAliasFinish.id,
        )
        .where(
            GContract.is_industry == false(),
            GRateScriptAliasFinish.finish_date != null(),
        )
        .order_by(GContract.name)
    )

    return render_template(
        "supplier_contracts.html",
        current_contracts=current_contracts,
        ended_contracts=ended_contracts,
    )


@gas.route("/supplier_contracts/<int:contract_id>")
def supplier_contract_get(contract_id):
    contract = GContract.get_supplier_by_id(g.sess, contract_id)
    rate_scripts = g.sess.execute(
        select(GRateScript)
        .where(GRateScript.g_contract == contract)
        .order_by(GRateScript.start_date.desc())
    ).scalars()

    now = Datetime.utcnow() - relativedelta(months=1)
    month_start = Datetime(now.year, now.month, 1)
    month_finish = month_start + relativedelta(months=1) - HH

    return render_template(
        "supplier_contract.html",
        contract=contract,
        month_start=month_start,
        month_finish=month_finish,
        rate_scripts=rate_scripts,
    )


@gas.route("/supplier_contracts/add")
def supplier_contract_add_get():
    contracts = g.sess.execute(
        select(GContract)
        .where(GContract.is_industry == false())
        .order_by(GContract.name)
    )
    return render_template("supplier_contract_add.html", contracts=contracts)


@gas.route("/supplier_contracts/add", methods=["POST"])
def supplier_contract_add_post():
    try:
        name = req_str("name")
        start_date = req_date("start")
        charge_script = req_str("charge_script")
        properties = req_zish("properties")

        contract = GContract.insert_supplier(
            g.sess, name, charge_script, properties, start_date, None, {}
        )
        g.sess.commit()
        return chellow_redirect(f"/supplier_contracts/{contract.id}", 303)

    except BadRequest as e:
        flash(e.description)
        g.sess.rollback()
        contracts = g.sess.query(GContract).order_by(GContract.name)
        return make_response(
            render_template("supplier_contract_add.html", contracts=contracts), 400
        )


@gas.route("/supplier_contracts/<int:g_contract_id>/edit")
def supplier_contract_edit_get(g_contract_id):
    g_contract = GContract.get_supplier_by_id(g.sess, g_contract_id)
    return render_template("supplier_contract_edit.html", g_contract=g_contract)


@gas.route("/supplier_contracts/<int:g_contract_id>/edit", methods=["POST"])
def supplier_contract_edit_post(g_contract_id):
    try:
        g_contract = GContract.get_by_id(g.sess, g_contract_id)
        name = req_str("name")
        charge_script = req_str("charge_script")
        properties = req_zish("properties")
        g_contract.update(name, charge_script, properties)
        g.sess.commit()
        return chellow_redirect(f"/supplier_contracts/{g_contract.id}", 303)
    except BadRequest as e:
        flash(e.description)
        g.sess.rollback()
        return make_response(
            render_template("supplier_contract_edit.html", g_contract=g_contract), 400
        )


@gas.route("/supplier_contracts/<int:g_contract_id>/edit", methods=["DELETE"])
def supplier_contract_edit_delete(g_contract_id):
    try:
        g_contract = GContract.get_by_id(g.sess, g_contract_id)
        g_contract.delete(g.sess)
        g.sess.commit()
        return hx_redirect("/supplier_contracts", 303)
    except BadRequest as e:
        flash(e.description)
        g.sess.rollback()
        return make_response(
            render_template("supplier_contract_edit.html", g_contract=g_contract), 400
        )


@gas.route("/supplier_contracts/<int:g_contract_id>/add_rate_script")
def supplier_rate_script_add_get(g_contract_id):
    now = utc_datetime_now()
    initial_date = utc_datetime(now.year, now.month)
    g_contract = GContract.get_by_id(g.sess, g_contract_id)
    return render_template(
        "supplier_rate_script_add.html",
        g_contract=g_contract,
        initial_date=initial_date,
    )


@gas.route("/supplier_contracts/<int:g_contract_id>/add_rate_script", methods=["POST"])
def supplier_rate_script_add_post(g_contract_id):
    try:
        g_contract = GContract.get_by_id(g.sess, g_contract_id)
        start_date = req_date("start")
        g_rate_script = g_contract.insert_g_rate_script(g.sess, start_date, {})
        g.sess.commit()
        return chellow_redirect(f"/supplier_rate_scripts/{g_rate_script.id}", 303)
    except BadRequest as e:
        flash(e.description)
        now = utc_datetime_now()
        initial_date = utc_datetime(now.year, now.month)
        return make_response(
            render_template(
                "supplier_rate_script_add.html",
                g_contract=g_contract,
                initial_date=initial_date,
            ),
            400,
        )


@gas.route("/supplier_rate_scripts/<int:g_rate_script_id>/edit")
def supplier_rate_script_edit_get(g_rate_script_id):
    g_rate_script = GRateScript.get_by_id(g.sess, g_rate_script_id)
    return render_template(
        "supplier_rate_script_edit.html", g_rate_script=g_rate_script
    )


@gas.route("/supplier_rate_scripts/<int:g_rate_script_id>/edit", methods=["POST"])
def supplier_rate_script_edit_post(g_rate_script_id):
    try:
        g_rate_script = GRateScript.get_by_id(g.sess, g_rate_script_id)
        g_contract = g_rate_script.g_contract
        if "delete" in request.values:
            g_contract.delete_g_rate_script(g.sess, g_rate_script)
            g.sess.commit()
            return chellow_redirect(f"/supplier_contracts/{g_contract.id}", 303)
        else:
            script = req_zish("script")
            start_date = req_date("start")
            has_finished = req_checkbox("has_finished")
            finish_date = req_date("finish") if has_finished else None
            g_contract.update_g_rate_script(
                g.sess, g_rate_script, start_date, finish_date, script
            )
            g.sess.commit()
            return chellow_redirect(f"/supplier_rate_scripts/{g_rate_script.id}", 303)
    except BadRequest as e:
        flash(e.description)
        g.sess.rollback()
        return make_response(
            render_template(
                "supplier_rate_script_edit.html", g_rate_script=g_rate_script
            ),
            400,
        )


@gas.route("/supplier_rate_scripts/<int:g_rate_script_id>")
def supplier_rate_script_get(g_rate_script_id):
    g_rate_script = GRateScript.get_by_id(g.sess, g_rate_script_id)
    g_contract = g_rate_script.g_contract
    next_g_rate_script = (
        g.sess.execute(
            select(GRateScript)
            .where(
                GRateScript.g_contract == g_contract,
                GRateScript.start_date > g_rate_script.start_date,
            )
            .order_by(GRateScript.start_date)
        )
        .scalars()
        .first()
    )
    previous_g_rate_script = (
        g.sess.execute(
            select(GRateScript)
            .where(
                GRateScript.g_contract == g_contract,
                GRateScript.start_date < g_rate_script.start_date,
            )
            .order_by(GRateScript.start_date.desc())
        )
        .scalars()
        .first()
    )
    return render_template(
        "supplier_rate_script.html",
        g_rate_script=g_rate_script,
        next_g_rate_script=next_g_rate_script,
        previous_g_rate_script=previous_g_rate_script,
    )



import csv
import decimal
from decimal import Decimal

from werkzeug.exceptions import BadRequest

from xlrd import open_workbook

from chellow.utils import HH, ct_datetime, to_utc, utc_datetime_now


def get_value(row, title_row, name):
    try:
        idx = title_row.index(name)
    except ValueError:
        return None

    try:
        val = row[idx].value
    except IndexError:
        raise BadRequest(
            f"For the name '{name}', the index is {idx} which is beyond the "
            f"end of the row."
        )

    if isinstance(val, str):
        return val.strip()
    else:
        return val


def get_dec(row, title_row, name):
    val = get_value(row, title_row, name)

    try:
        return val if val is None else Decimal(str(val))
    except decimal.InvalidOperation:
        return None


def get_rate(row, title_row, name):
    raw = get_dec(row, title_row, name)
    return None if raw is None else [raw / Decimal("100")]


def get_date_ct(row, title_row, name):
    s = get_value(row, title_row, name)
    nums = list(map(int, s.split("-")))
    return ct_datetime(nums[0], nums[1], nums[2])


ELEMENT_NAME_LOOKUP = {
    "LVY-CMLOB-ALL": "capmechob",
    "LVY-CMLOP-ALL": "capmechop",
    "DUOS-STND-SITE": "duos-fixed",
    "DUOS-AVAIL-AV": "duos-availability",
    "DUOS-UNIT-AMBER": "duos-amber",
    "DUOS-UNIT-GREEN": "duos-green",
    "DUOS-UNIT-RED": "duos-red",
    "LVY-AAHEDC-ALL": "aahedc",
    "LVY-BSUOS-ALL": "bsuos",
    "LVY-CCL-ALL": "ccl",
    "LVY-CFDOB-ALL": "cfdob",
    "LVY-CFDOP-ALL": "cfdop",
    "LVY-FIT-ALL": "fit",
    "LVY-RO-ALL": "ro",
    "NRG-UNIT-SUMMER": "summer",
    "NRG-UNIT-WINTER": "winter",
    "SYS-ADMIN-ALL": "admin",
    "TUOS-TRIAD-SITE": "triad",
}


def _parse_row(issue_date, row, row_index, datemode, title_row):
    titles = [c.value for c in title_row if c.value is not None]
    bill_start_date_ct = get_date_ct(row, titles, "BILL START DATE")
    start_date = to_utc(bill_start_date_ct)
    bill_finish_date_ct = get_date_ct(row, titles, "BILL END DATE") + (HH * 47)
    finish_date = to_utc(bill_finish_date_ct)

    elements = {}

    kwh = 0
    net_gbp = Decimal("0.00")
    vat_gbp = Decimal("0.00")
    for title in titles[3:]:
        val = get_dec(row, titles, title)
        if val is not None:
            parts = title.split()
            if len(parts) != 3:
                continue
            element_code, typ, units = parts
            element_name = ELEMENT_NAME_LOOKUP[element_code]
            try:
                element = elements[element_name]
            except KeyError:
                element = elements[element_name] = {
                    "name": element_name,
                    "breakdown": {},
                    "net": Decimal("0.00"),
                    "start_date": start_date,
                    "finish_date": finish_date,
                }

            bd = element["breakdown"]
            if typ == "COST":
                element["net"] += round(val, 2)
                net_gbp += round(val, 2)
            elif typ == "USE":
                bd[units.lower()] = val
                if element_name == "aahedc":
                    kwh += val
            elif typ == "RATE":
                bd["rate"] = val

    bd = {"raw_lines": [str(title_row), str(row)]}

    reference = f"{issue_date.strftime('%Y%m%dT%H%M')}_{row_index + 1}"

    return {
        "bill_type_code": "N",
        "kwh": kwh,
        "vat": Decimal("0.00"),
        "net": net_gbp,
        "gross": net_gbp + vat_gbp,
        "reads": [],
        "breakdown": {},
        "account": "2007",
        "issue_date": issue_date,
        "start_date": start_date,
        "finish_date": finish_date,
        "mpan_core": "22 0003 0354 632",
        "reference": reference,
        "elements": list(elements.values()),
    }


class Parser:
    def __init__(self, f):
        self.book = open_workbook(file_contents=f.read())
        self.sheet = self.book.sheet_by_index(1)

        self.last_line = None
        lines = (self._set_last_line(i, l) for i, l in enumerate(f))
        self.reader = csv.reader(lines, skipinitialspace=True)
        self._line_number = None
        self._title_line = None

    @property
    def line_number(self):
        return None if self._line_number is None else self._line_number + 1

    def _set_last_line(self, i, line):
        self._line_numer = i
        self.last_line = line
        if i == 0:
            self._title_line = line
        return line

    def make_raw_bills(self):
        bills = []
        title_row = self.sheet.row(0)
        issue_date = utc_datetime_now()
        for row_index in range(1, self.sheet.nrows):
            row = self.sheet.row(row_index)
            datemode = self.book.datemode
            try:
                bills.append(
                    _parse_row(issue_date, row, row_index, datemode, title_row)
                )
            except BadRequest as e:
                raise BadRequest(f"On row {row_index}: {e.description}")
        return bills



from decimal import Decimal
from io import BytesIO

from chellow.e.hh_parser_df2 import create_parser
from chellow.utils import utc_datetime


def test_parser():
    reader_lines = [
        "#F2",
        "#O 2248274887874-jlkkk38",
        "#S 1",
        "16/04/2024,00:30,0.77,A",
        "16/04/2024,01:00,0.064,A",
        "16/04/2024,01:30,0,A",
    ]
    reader = BytesIO("\n".join(reader_lines).encode("utf8"))
    mpan_map = {}
    messages = []

    parser = create_parser(reader, mpan_map, messages)
    actuals = list(parser)
    expected = [
        {
            "mpan_core": "22 4827 4887 874",
            "channel_type": "ACTIVE",
            "start_date": utc_datetime(2024, 4, 16, 0, 0),
            "value": Decimal("0.77"),
            "status": "A",
        },
        {
            "mpan_core": "22 4827 4887 874",
            "channel_type": "ACTIVE",
            "start_date": utc_datetime(2024, 4, 16, 0, 30),
            "value": Decimal("0.064"),
            "status": "A",
        },
        {
            "mpan_core": "22 4827 4887 874",
            "channel_type": "ACTIVE",
            "start_date": utc_datetime(2024, 4, 16, 1, 0),
            "value": Decimal("0"),
            "status": "A",
        },
    ]

    assert actuals == expected



import atexit
import collections
import threading
import traceback
from datetime import timedelta
from importlib import import_module

import requests

from werkzeug.exceptions import BadRequest

from chellow.models import (
    Contract,
    Session,
)
from chellow.utils import ct_datetime_now, hh_format, utc_datetime_now

importer = None


def download_file(log, scripting_key, file_name):
    params = {"key": scripting_key}
    url = f"https://downloads.elexonportal.co.uk/file/download/{file_name}"

    log(f"Downloading from {url}?key={scripting_key}")

    return requests.get(url, params=params, timeout=120)


def run_import(sess, log, set_progress, scripting_key):
    log("Starting to import data from Elexon")

    for mod_name in (
        "chellow.e.system_price",
        "chellow.e.tlms",
        "chellow.e.rcrc",
        "chellow.e.lafs",
    ):
        mod = import_module(mod_name)
        try:
            mod.elexon_import(sess, log, set_progress, scripting_key)
        except TypeError as e:
            raise BadRequest(f"Problem with module {mod_name}: {e}") from e


ELEXON_PORTAL_SCRIPTING_KEY_KEY = "elexonportal_scripting_key"
LAST_RUN_KEY = "last_run"
GLOBAL_ALERT = "There's a problem with an <a href='/e/elexon'>Elexon import</a>."
ELEXON_STATE_KEY = "elexon"


class Elexon(threading.Thread):
    def __init__(self):
        super().__init__(name="Elexon")
        self.messages = collections.deque(maxlen=500)
        self.progress = ""
        self.stopped = threading.Event()
        self.going = threading.Event()
        self.global_alert = None

    def stop(self):
        self.stopped.set()
        self.going.set()
        self.join()

    def go(self):
        self.going.set()

    def log(self, message):
        self.messages.appendleft(
            f"{ct_datetime_now().strftime('%Y-%m-%d %H:%M:%S')} - {message}"
        )

    def set_progress(self, progress):
        self.progress = progress

    def run(self):
        while not self.stopped.is_set():
            with Session() as sess:
                try:
                    config = Contract.get_non_core_by_name(sess, "configuration")
                    state = config.make_state()
                    elexon_state = state.get(ELEXON_STATE_KEY, {})
                except BaseException as e:
                    msg = f"{e.description} " if isinstance(e, BadRequest) else ""
                    self.log(f"{msg}{traceback.format_exc()}")
                    self.global_alert = GLOBAL_ALERT
                    sess.rollback()

            last_run = elexon_state.get(LAST_RUN_KEY)
            if last_run is None or utc_datetime_now() - last_run > timedelta(days=1):
                self.going.set()

            if self.going.is_set():
                self.global_alert = None
                with Session() as sess:
                    try:
                        config = Contract.get_non_core_by_name(sess, "configuration")
                        state = config.make_state()
                        try:
                            elexon_state = state[ELEXON_STATE_KEY]
                        except KeyError:
                            elexon_state = state[ELEXON_STATE_KEY] = {}

                        elexon_state[LAST_RUN_KEY] = utc_datetime_now()
                        config.update_state(state)
                        props = config.make_properties()
                        scripting_key = props.get(ELEXON_PORTAL_SCRIPTING_KEY_KEY)
                        sess.commit()
                        if scripting_key is None:
                            raise BadRequest(
                                f"The property {ELEXON_PORTAL_SCRIPTING_KEY_KEY} "
                                f"cannot be found in the configuration properties. "
                                f"A scripting key can be obtained from "
                                f"https://www.elexonportal.co.uk/"
                            )
                        run_import(sess, self.log, self.set_progress, scripting_key)
                    except BaseException as e:
                        msg = f"{e.description} " if isinstance(e, BadRequest) else ""
                        self.log(f"{msg}{traceback.format_exc()}")
                        self.global_alert = GLOBAL_ALERT
                        sess.rollback()
                    finally:
                        self.going.clear()
                        self.log("Finished importing Elexon data.")

            else:
                self.log(
                    f"The importer was last run at {hh_format(last_run)}. There will "
                    f"be another import when 24 hours have elapsed since the last run."
                )
                self.going.wait(60 * 60)


def get_importer():
    return importer


def startup():
    global importer
    importer = Elexon()
    importer.start()


@atexit.register
def shutdown():
    if importer is not None:
        importer.stop()



from datetime import datetime as Datetime
from decimal import Decimal
from io import BytesIO

from openpyxl import Workbook

from chellow.e.bill_parsers.amr_dc_stark_xlsx import Parser
from chellow.utils import ct_datetime, to_utc


def test_make_bills(mocker, sess):
    f = BytesIO()
    wb = Workbook()
    wb.create_sheet(title="HH", index=1)
    sheet = wb.worksheets[1]
    sheet.insert_rows(0, 12)
    sheet.insert_cols(0, 13)

    sheet["C6"] = Datetime(2025, 10, 30)

    sheet["B11"] = "MPAN"
    sheet["C11"] = "Site Address"
    sheet["D11"] = "Comms"
    sheet["E11"] = "Start Date"
    sheet["F11"] = "Type"
    sheet["G11"] = "Period From"
    sheet["H11"] = "Period To"
    sheet["I11"] = "DC PA"
    sheet["J11"] = "Charge"
    sheet["K11"] = "Est. VAT"
    sheet["L11"] = "Est. Total"

    sheet["B12"] = "2276173918389"
    sheet["C12"] = "Water Works"
    sheet["D12"] = "GPRS"
    sheet["E12"] = ""
    sheet["F12"] = "Settled"
    sheet["G12"] = Datetime(2025, 7, 1)
    sheet["H12"] = Datetime(2025, 9, 30)
    sheet["I12"] = Decimal("16.7")
    sheet["J12"] = Decimal("5.82")
    sheet["K12"] = Decimal("1.57")
    sheet["L12"] = Decimal("8.45")

    wb.save(f)
    f.seek(0)
    p = Parser(f)
    result = p.make_raw_bills()

    expected = [
        {
            "bill_type_code": "N",
            "kwh": Decimal("0"),
            "vat": Decimal("1.57"),
            "net": Decimal("5.82"),
            "gross": Decimal("8.45"),
            "reads": [],
            "breakdown": {
                "raw_lines": [],
                "vat": {
                    20: {
                        "net": Decimal("5.82"),
                        "vat": Decimal("1.57"),
                    },
                },
            },
            "account": "22 7617 3918 389",
            "issue_date": to_utc(ct_datetime(2025, 10, 30)),
            "start_date": to_utc(ct_datetime(2025, 7, 1)),
            "finish_date": to_utc(ct_datetime(2025, 9, 30, 23, 30)),
            "mpan_core": "22 7617 3918 389",
            "reference": "20250701_20250930_20251030_22 7617 3918 389",
            "elements": [
                {
                    "name": "mpan",
                    "start_date": to_utc(ct_datetime(2025, 7, 1)),
                    "finish_date": to_utc(ct_datetime(2025, 9, 30, 23, 30)),
                    "net": Decimal("5.82"),
                    "breakdown": {
                        "rate": {Decimal("16.7")},
                        "days": 92,
                        "settlement-status": {"settlement"},
                        "comm": {"GPRS"},
                    },
                }
            ],
        }
    ]

    assert result == expected



import csv
import sys
import threading
import traceback

from flask import g, redirect, render_template, request

from sqlalchemy import Integer, any_, cast, select
from sqlalchemy.dialects.postgresql import JSONB, array

from chellow.dloads import open_file
from chellow.gas.issues import make_issue_bundles
from chellow.models import (
    GIssue,
    RSession,
    User,
)
from chellow.utils import (
    csv_make_val,
    req_checkbox,
)


def _make_bundles(sess, contract_ids, owner_ids, supply_ids, is_opens, limit=None):
    q = select(GIssue).order_by(GIssue.is_open.desc(), GIssue.date_created)
    if limit is not None:
        q = q.limit(limit)
    if len(contract_ids) > 0:
        q = q.where(GIssue.g_contract_id.in_(contract_ids))
    if len(owner_ids) > 0:
        q = q.where(cast(GIssue.properties["owner_id"], Integer).in_(owner_ids))
    if len(supply_ids) > 0:
        ids_jsonb = array([cast([sid], JSONB) for sid in supply_ids])
        q = q.where(GIssue.properties["supply_ids"].op("@>")(any_(ids_jsonb)))
    if len(is_opens) > 0:
        q = q.where(GIssue.is_open.in_(is_opens))
    issues = sess.scalars(q)
    return make_issue_bundles(sess, issues)


def _make_vals(sess, contract_ids, owner_ids, supply_ids, is_opens):
    for bundle in _make_bundles(sess, contract_ids, owner_ids, supply_ids, is_opens):
        issue = bundle["issue"]
        props = issue.properties
        owner = bundle["owner"]
        contract = issue.g_contract
        values = {
            "issue_id": issue.id,
            "contract_name": contract.name,
            "date_created": issue.date_created,
            "owner": None if owner is None else owner.email_address,
            "status": "open" if issue.is_open else "closed",
            "subject": props.get("subject"),
            "mprn_core": None,
            "site_code": None,
            "site_name": None,
        }
        latest_entry = bundle["latest_entry"]
        if latest_entry is None:
            values["latest_entry_timestamp"] = None
            values["latest_entry_markdown"] = None
        else:
            values["latest_entry_timestamp"] = latest_entry.timestamp
            values["latest_entry_markdown"] = latest_entry.markdown

        supplies = bundle["supplies"]
        if len(supplies) == 0:
            yield values
        else:
            for supply_bundle in supplies:
                values = values.copy()
                site = supply_bundle["site"]
                values["mprn"] = supply_bundle["supply"].mprn
                values["site_code"] = site.code
                values["site_name"] = site.name
                yield values


def content(user_id, contract_ids, owner_ids, supply_ids, is_opens):
    f = writer = None
    try:
        with RSession() as sess:
            user = User.get_by_id(sess, user_id)
            f = open_file("issues.csv", user, mode="w", newline="")
            writer = csv.writer(f, lineterminator="\n")
            titles = (
                "issue_id",
                "contract_name",
                "date_created",
                "owner",
                "status",
                "subject",
                "mprn",
                "site_code",
                "site_name",
                "latest_entry_timestamp",
                "latest_entry_markdown",
            )
            writer.writerow(titles)
            for vals in _make_vals(sess, contract_ids, owner_ids, supply_ids, is_opens):
                writer.writerow(csv_make_val(vals[t]) for t in titles)

    except BaseException:
        msg = traceback.format_exc()
        sys.stderr.write(msg)
        writer.writerow([msg])
    finally:
        if f is not None:
            f.close()


LIMIT = 200


def do_get(sess):
    contract_ids = [int(x) for x in request.values.getlist("contract_id")]
    owner_ids = [int(x) for x in request.values.getlist("owner_id")]
    supply_ids = [int(x) for x in request.values.getlist("supply_id")]
    is_opens = [x == "true" for x in request.values.getlist("is_open")]
    as_csv = req_checkbox("as_csv")

    if as_csv:
        args = (g.user.id, contract_ids, owner_ids, supply_ids, is_opens)
        threading.Thread(target=content, args=args).start()
        return redirect("/downloads", 303)
    else:
        issue_bundles = _make_bundles(
            sess, contract_ids, owner_ids, supply_ids, is_opens, limit=LIMIT
        )
        return render_template(
            "reports/g_issues.html",
            contract_ids=contract_ids,
            owner_ids=owner_ids,
            supply_ids=supply_ids,
            is_opens=is_opens,
            limit=LIMIT,
            issue_bundles=issue_bundles,
        )
